# Copyright (c) 2017 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import mock

from sahara.service.edp.job_binaries import base as jb_base
from sahara.tests.unit import base


class _FakeJobBinary(jb_base.JobBinaryType):
    def copy_binary_to_cluster(self, job_binary, **kwargs):
        return 'valid path'


class JobBinaryManagerSupportTest(base.SaharaTestCase):

    def setUp(self):
        super(JobBinaryManagerSupportTest, self).setUp()
        self.job_binary = _FakeJobBinary()

    def test_generate_valid_path(self):
        jb = mock.Mock()
        jb.name = 'jb_name.jar'
        res = self.job_binary._generate_valid_path(jb)
        self.assertEqual('/tmp/jb_name.jar', res)



# Copyright (c) 2015 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from castellan import options as castellan
from oslo_config import cfg

from sahara.utils.openstack import base as utils


opts = [
    cfg.BoolOpt('use_barbican_key_manager', default=False,
                help='Enable the usage of the OpenStack Key Management '
                     'service provided by barbican.'),
]

castellan_opts = [
    cfg.StrOpt('barbican_api_endpoint',
               help='The endpoint to use for connecting to the barbican '
                    'api controller. By default, castellan will use the '
                    'URL from the service catalog.'),
    cfg.StrOpt('barbican_api_version', default='v1',
               help='Version of the barbican API, for example: "v1"'),
]

castellan_group = cfg.OptGroup(name='castellan',
                               title='castellan key manager options')

CONF = cfg.CONF
CONF.register_group(castellan_group)
CONF.register_opts(opts)
CONF.register_opts(castellan_opts, group=castellan_group)


def validate_config():
    if CONF.use_barbican_key_manager:
        # NOTE (elmiko) there is no need to set the api_class as castellan
        # uses barbican by default.
        castellan.set_defaults(CONF, auth_endpoint=utils.retrieve_auth_url())
    else:
        castellan.set_defaults(CONF, api_class='sahara.service.castellan.'
                               'sahara_key_manager.SaharaKeyManager')



# Copyright (c) 2016 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_utils import uuidutils

import abc
import collections
import copy
import functools
import itertools
from os import path

import jsonschema
import six
import yaml

from sahara import exceptions as ex
from sahara.i18n import _
from sahara.plugins import exceptions as p_ex
from sahara.plugins import utils


def transform_exception(from_type, to_type, transform_func=None):
    """Decorator to transform exception types.

    :param from_type: The type of exception to catch and transform.
    :param to_type: The type of exception to raise instead.
    :param transform_func: A function to transform from_type into
        to_type, which must be of the form func(exc, to_type).
        Defaults to:
        lambda exc, new_type: new_type(exc.message)
    """
    if not transform_func:
        transform_func = lambda exc, new_type: new_type(exc.message)

    def decorator(func):
        @functools.wraps(func)
        def handler(*args, **kwargs):
            try:
                func(*args, **kwargs)
            except from_type as exc:
                raise transform_func(exc, to_type)
        return handler
    return decorator


def validate_instance(instance, validators, test_only=False, **kwargs):
    """Runs all validators against the specified instance.

    :param instance: An instance to validate.
    :param validators: A sequence of ImageValidators.
    :param test_only: If true, all validators will only verify that a
        desired state is present, and fail if it is not. If false, all
        validators will attempt to enforce the desired state if possible,
        and succeed if this enforcement succeeds.
    :raises ImageValidationError: If validation fails.
    """
    with instance.remote() as remote:
        for validator in validators:
            validator.validate(remote, test_only=test_only, **kwargs)


class ImageArgument(object):
    """An argument used by an image manifest."""

    SPEC_SCHEMA = {
        "type": "object",
        "items": {
            "type": "object",
            "properties": {
                "target_variable": {
                    "type": "string",
                    "minLength": 1
                },
                "description": {
                    "type": "string",
                    "minLength": 1
                },
                "default": {
                    "type": "string",
                    "minLength": 1
                },
                "required": {
                    "type": "boolean",
                    "minLength": 1
                },
                "choices": {
                    "type": "array",
                    "minLength": 1,
                    "items": {
                        "type": "string"
                    }
                }
            }
        }
    }

    @classmethod
    def from_spec(cls, spec):
        """Constructs and returns a set of arguments from a specification.

        :param spec: The specification for the argument set.
        :return: A dict of arguments built to the specification.
        """

        jsonschema.validate(spec, cls.SPEC_SCHEMA)
        arguments = {name: cls(name,
                               arg.get('description'),
                               arg.get('default'),
                               arg.get('required'),
                               arg.get('choices'))
                     for name, arg in six.iteritems(spec)}
        reserved_names = ['distro', 'test_only']
        for name, arg in six.iteritems(arguments):
            if name in reserved_names:
                raise p_ex.ImageValidationSpecificationError(
                    _("The following argument names are reserved: "
                      "{names}").format(reserved_names))
            if not arg.default and not arg.required:
                raise p_ex.ImageValidationSpecificationError(
                    _("Argument {name} is not required and must specify a "
                      "default value.").format(name=arg.name))
            if arg.choices and arg.default and arg.default not in arg.choices:
                raise p_ex.ImageValidationSpecificationError(
                    _("Argument {name} specifies a default which is not one "
                      "of its choices.").format(name=arg.name))
        return arguments

    def __init__(self, name, description=None, default=None, required=False,
                 choices=None):
        self.name = name
        self.description = description
        self.default = default
        self.required = required
        self.choices = choices


@six.add_metaclass(abc.ABCMeta)
class ImageValidator(object):
    """Validates the image spawned to an instance via a set of rules."""

    @abc.abstractmethod
    def validate(self, remote, test_only=False, **kwargs):
        """Validates the image.

        :param remote: A remote socket to the instance.
        :param test_only: If true, all validators will only verify that a
            desired state is present, and fail if it is not. If false, all
            validators will attempt to enforce the desired state if possible,
            and succeed if this enforcement succeeds.
        :raises ImageValidationError: If validation fails.
        """
        pass


@six.add_metaclass(abc.ABCMeta)
class SaharaImageValidatorBase(ImageValidator):
    """Base class for Sahara's native image validation."""

    DISTRO_KEY = 'distro'
    TEST_ONLY_KEY = 'test_only'

    ORDERED_VALIDATORS_SCHEMA = {
        "type": "array",
        "items": {
            "type": "object",
            "minProperties": 1,
            "maxProperties": 1
        }
    }

    _DISTRO_FAMILES = {
        'centos': 'redhat',
        'centos7': 'redhat',
        'fedora': 'redhat',
        'redhat': 'redhat',
        'rhel': 'redhat',
        'redhatenterpriseserver': 'redhat',
        'ubuntu': 'debian'
    }

    @staticmethod
    def get_validator_map(custom_validator_map=None):
        """Gets the map of validator name token to validator class.

        :param custom_validator_map: A map of validator names and classes to
            add to the ones Sahara provides by default. These will take
            precedence over the base validators in case of key overlap.
        :return: A map of validator names and classes.
        """
        default_validator_map = {
            'package': SaharaPackageValidator,
            'script': SaharaScriptValidator,
            'copy_script': SaharaCopyScriptValidator,
            'any': SaharaAnyValidator,
            'all': SaharaAllValidator,
            'os_case': SaharaOSCaseValidator,
            'argument_case': SaharaArgumentCaseValidator,
            'argument_set': SaharaArgumentSetterValidator,
        }
        if custom_validator_map:
            default_validator_map.update(custom_validator_map)
        return default_validator_map

    @classmethod
    def from_yaml(cls, yaml_path, validator_map=None, resource_roots=None,
                  package='sahara'):
        """Constructs and returns a validator from the provided yaml file.

        :param yaml_path: The relative path to a yaml file.
        :param validator_map: A map of validator name to class.
        :param resource_roots: The roots from which relative paths to
            resources (scripts and such) will be referenced. Any resource will
            be pulled from the first path in the list at which a file exists.
        :return: A SaharaImageValidator built to the yaml specification.
        """
        validator_map = validator_map or {}
        resource_roots = resource_roots or []
        file_text = utils.get_file_text(yaml_path, package)
        spec = yaml.safe_load(file_text)
        validator_map = cls.get_validator_map(validator_map)
        return cls.from_spec(spec, validator_map, resource_roots, package)

    @classmethod
    def from_spec(cls, spec, validator_map, resource_roots, package='sahara'):
        """Constructs and returns a validator from a specification object.

        :param spec: The specification for the validator.
        :param validator_map: A map of validator name to class.
        :param resource_roots: The roots from which relative paths to
            resources (scripts and such) will be referenced. Any resource will
            be pulled from the first path in the list at which a file exists.
        :return: A validator built to the specification.
        """
        pass

    @classmethod
    def from_spec_list(cls, specs, validator_map, resource_roots,
                       package='sahara'):
        """Constructs a list of validators from a list of specifications.

        :param specs: A list of validator specifications, each of which
            will be a dict of size 1, where the key represents the validator
            type and the value respresents its specification.
        :param validator_map: A map of validator name to class.
        :param resource_roots: The roots from which relative paths to
            resources (scripts and such) will be referenced. Any resource will
            be pulled from the first path in the list at which a file exists.
        :return: A list of validators.
        """
        validators = []
        for spec in specs:
            validator_class, validator_spec = cls.get_class_from_spec(
                spec, validator_map)
            validators.append(validator_class.from_spec(
                validator_spec, validator_map, resource_roots, package))
        return validators

    @classmethod
    def get_class_from_spec(cls, spec, validator_map):
        """Gets the class and specification from a validator dict.

        :param spec: A validator specification including its type: a dict of
            size 1, where the key represents the validator type and the value
            respresents its configuration.
        :param validator_map: A map of validator name to class.
        :return: A tuple of validator class and configuration.
        """
        key, value = list(six.iteritems(spec))[0]
        validator_class = validator_map.get(key, None)
        if not validator_class:
            raise p_ex.ImageValidationSpecificationError(
                _("Validator type %s not found.") % validator_class)
        return validator_class, value

    class ValidationAttemptFailed(object):
        """An object representing a failed validation attempt.

        Primarily for use by the SaharaAnyValidator, which must aggregate
        failures for error exposition purposes.
        """
        def __init__(self, exception):
            self.exception = exception

        def __bool__(self):
            return False

        def __nonzero__(self):
            return False

    def try_validate(self, remote, test_only=False,
                     image_arguments=None, **kwargs):
        """Attempts to validate, but returns rather than raising on failure.

        :param remote: A remote socket to the instance.
        :param test_only: If true, all validators will only verify that a
            desired state is present, and fail if it is not. If false, all
            validators will attempt to enforce the desired state if possible,
            and succeed if this enforcement succeeds.
        :param image_arguments: A dictionary of image argument values keyed by
            argument name.
        :return: True if successful, ValidationAttemptFailed object if failed.
        """
        try:
            self.validate(
                remote, test_only=test_only,
                image_arguments=image_arguments, **kwargs)
            return True
        except p_ex.ImageValidationError as exc:
            return self.ValidationAttemptFailed(exc)


class SaharaImageValidator(SaharaImageValidatorBase):
    """The root of any tree of SaharaImageValidators.

    This validator serves as the root of the tree for SaharaImageValidators,
    and provides any needed initialization (such as distro retrieval.)
    """

    SPEC_SCHEMA = {
        "title": "SaharaImageValidator",
        "type": "object",
        "properties": {
            "validators": SaharaImageValidatorBase.ORDERED_VALIDATORS_SCHEMA
        },
        "required": ["validators"]
    }

    def get_argument_list(self):
        return [argument for name, argument
                in six.iteritems(self.arguments)]

    @classmethod
    def from_spec(cls, spec, validator_map, resource_roots, package='sahara'):
        """Constructs and returns a validator from a specification object.

        :param spec: The specification for the validator: a dict containing
            the key "validators", which contains a list of validator
            specifications.
        :param validator_map: A map of validator name to class.
        :param resource_roots: The roots from which relative paths to
            resources (scripts and such) will be referenced. Any resource will
            be pulled from the first path in the list at which a file exists.
        :return: A SaharaImageValidator containing all specified validators.
        """
        jsonschema.validate(spec, cls.SPEC_SCHEMA)
        arguments_spec = spec.get('arguments', {})
        arguments = ImageArgument.from_spec(arguments_spec)
        validators_spec = spec['validators']
        validator = SaharaAllValidator.from_spec(
            validators_spec, validator_map, resource_roots, package)
        return cls(validator, arguments)

    def __init__(self, validator, arguments):
        """Constructor method.

        :param validator: A SaharaAllValidator containing the specified
            validators.
        """
        self.validator = validator
        self.validators = validator.validators
        self.arguments = arguments

    @transform_exception(ex.RemoteCommandException, p_ex.ImageValidationError)
    def validate(self, remote, test_only=False,
                 image_arguments=None, **kwargs):
        """Attempts to validate the image.

        Before deferring to contained validators, performs one-time setup
        steps such as distro discovery.

        :param remote: A remote socket to the instance.
        :param test_only: If true, all validators will only verify that a
            desired state is present, and fail if it is not. If false, all
            validators will attempt to enforce the desired state if possible,
            and succeed if this enforcement succeeds.
        :param image_arguments: A dictionary of image argument values keyed by
            argument name.
        :raises ImageValidationError: If validation fails.
        """
        argument_values = {}
        for name, argument in six.iteritems(self.arguments):
            if name not in image_arguments:
                if argument.required:
                    raise p_ex.ImageValidationError(
                        _("Argument {name} is required for image "
                          "processing.").format(name=name))
                else:
                    argument_values[name] = argument.default
            else:
                value = image_arguments[name]
                choices = argument.choices
                if choices and value not in choices:
                    raise p_ex.ImageValidationError(
                        _("Value for argument {name} must be one of "
                          "{choices}.").format(name=name, choices=choices))
                else:
                    argument_values[name] = value
        argument_values[self.DISTRO_KEY] = remote.get_os_distrib()
        self.validator.validate(remote, test_only=test_only,
                                image_arguments=argument_values)


class SaharaPackageValidator(SaharaImageValidatorBase):
    """A validator that checks package installation state on the instance."""

    class Package(object):

        def __init__(self, name, version=None):
            self.name = name
            self.version = version

        def __str__(self):
            return ("%s-%s" % (self.name, self.version)
                    if self.version else self.name)

    _SINGLE_PACKAGE_SCHEMA = {
        "oneOf": [
            {
                "type": "object",
                "minProperties": 1,
                "maxProperties": 1,
                "additionalProperties": {
                    "type": "object",
                    "properties": {
                        "version": {
                            "type": "string",
                            "minLength": 1
                        },
                    }
                },
            },
            {
                "type": "string",
                "minLength": 1
            }
        ]
    }

    SPEC_SCHEMA = {
        "title": "SaharaPackageValidator",
        "oneOf": [
            _SINGLE_PACKAGE_SCHEMA,
            {
                "type": "array",
                "items": _SINGLE_PACKAGE_SCHEMA,
                "minLength": 1
            }
        ]
    }

    @classmethod
    def _package_from_spec(cls, spec):
        """Builds a single package object from a specification.

        :param spec: May be a string or single-length dictionary of name to
            configuration values.
        :return: A package object.
        """
        if isinstance(spec, six.string_types):
            return cls.Package(spec, None)
        else:
            package, properties = list(six.iteritems(spec))[0]
            version = properties.get('version', None)
            return cls.Package(package, version)

    @classmethod
    def from_spec(cls, spec, validator_map, resource_roots, package='sahara'):
        """Builds a package validator from a specification.

        :param spec: May be a string, a single-length dictionary of name to
            configuration values, or a list containing any number of either or
            both of the above. Configuration values may include:
            version: The version of the package to check and/or install.
        :param validator_map: A map of validator name to class.
        :param resource_roots: The roots from which relative paths to
            resources (scripts and such) will be referenced. Any resource will
            be pulled from the first path in the list at which a file exists.
        :return: A validator that will check that the specified package or
            packages are installed.
        """
        jsonschema.validate(spec, cls.SPEC_SCHEMA)
        packages = ([cls._package_from_spec(package_spec)
                     for package_spec in spec]
                    if isinstance(spec, list)
                    else [cls._package_from_spec(spec)])
        return cls(packages)

    def __init__(self, packages):
        self.packages = packages

    @transform_exception(ex.RemoteCommandException, p_ex.ImageValidationError)
    def validate(self, remote, test_only=False,
                 image_arguments=None, **kwargs):
        """Attempts to validate package installation on the image.

        Even if test_only=False, attempts to verify previous package
        installation offline before using networked tools to validate or
        install new packages.

        :param remote: A remote socket to the instance.
        :param test_only: If true, all validators will only verify that a
            desired state is present, and fail if it is not. If false, all
            validators will attempt to enforce the desired state if possible,
            and succeed if this enforcement succeeds.
        :param image_arguments: A dictionary of image argument values keyed by
            argument name.
        :raises ImageValidationError: If validation fails.
        """
        env_distro = image_arguments[self.DISTRO_KEY]
        env_family = self._DISTRO_FAMILES[env_distro]
        check, install = self._DISTRO_TOOLS[env_family]
        if not env_family:
            raise p_ex.ImageValidationError(
                _("Unknown distro: cannot verify or install packages."))
        try:
            check(self, remote)
        except (ex.SubprocessException, ex.RemoteCommandException,
                RuntimeError):
            if not test_only:
                install(self, remote)
                check(self, remote)
            else:
                raise

    def _dpkg_check(self, remote):
        check_cmd = ("dpkg -s %s" %
                     " ".join(str(package) for package in self.packages))
        return _sudo(remote, check_cmd)

    def _rpm_check(self, remote):
        check_cmd = ("rpm -q %s" %
                     " ".join(str(package) for package in self.packages))
        return _sudo(remote, check_cmd)

    def _yum_install(self, remote):
        install_cmd = (
            "yum install -y %s" %
            " ".join(str(package) for package in self.packages))
        _sudo(remote, install_cmd)

    def _apt_install(self, remote):
        install_cmd = (
            "DEBIAN_FRONTEND=noninteractive apt-get -y install %s" %
            " ".join(str(package) for package in self.packages))
        return _sudo(remote, install_cmd)

    _DISTRO_TOOLS = {
        "redhat": (_rpm_check, _yum_install),
        "debian": (_dpkg_check, _apt_install)
    }


class SaharaScriptValidator(SaharaImageValidatorBase):
    """A validator that runs a script on the instance."""

    _DEFAULT_ENV_VARS = [SaharaImageValidatorBase.TEST_ONLY_KEY,
                         SaharaImageValidatorBase.DISTRO_KEY]

    SPEC_SCHEMA = {
        "title": "SaharaScriptValidator",
        "oneOf": [
            {
                "type": "object",
                "minProperties": 1,
                "maxProperties": 1,
                "additionalProperties": {
                    "type": "object",
                    "properties": {
                        "env_vars": {
                            "type": "array",
                            "items": {
                                "type": "string"
                            }
                        },
                        "output": {
                            "type": "string",
                            "minLength": 1
                        },
                        "inline": {
                            "type": "string",
                            "minLength": 1
                        }
                    },
                }
            },
            {
                "type": "string"
            }
        ]
    }

    @classmethod
    def from_spec(cls, spec, validator_map, resource_roots, package='sahara'):
        """Builds a script validator from a specification.

        :param spec: May be a string or a single-length dictionary of name to
            configuration values. Configuration values include:
            env_vars: A list of environment variable names to send to the
                script.
            output: A key into which to put the stdout of the script in the
                image_arguments of the validation run.
        :param validator_map: A map of validator name to class.
        :param resource_roots: The roots from which relative paths to
            resources (scripts and such) will be referenced. Any resource will
            be pulled from the first path in the list at which a file exists.
        :return: A validator that will run a script on the image.
        """
        jsonschema.validate(spec, cls.SPEC_SCHEMA)

        script_contents = None
        if isinstance(spec, six.string_types):
            script_path = spec
            env_vars, output_var = cls._DEFAULT_ENV_VARS, None
        else:
            script_path, properties = list(six.iteritems(spec))[0]
            env_vars = cls._DEFAULT_ENV_VARS + properties.get('env_vars', [])
            output_var = properties.get('output', None)
            script_contents = properties.get('inline')

        if not script_contents:
            for root in resource_roots:
                file_path = path.join(root, script_path)
                script_contents = utils.try_get_file_text(file_path, package)
                if script_contents:
                    break

        if not script_contents:
            raise p_ex.ImageValidationSpecificationError(
                _("Script %s not found in any resource roots.") % script_path)

        return SaharaScriptValidator(script_contents, env_vars, output_var)

    def __init__(self, script_contents, env_vars=None, output_var=None):
        """Constructor method.

        :param script_contents: A string representation of the script.
        :param env_vars: A list of environment variables to send to the
            script.
        :param output_var: A key into which to put the stdout of the script in
            the image_arguments of the validation run.
        :return: A SaharaScriptValidator.
        """
        self.script_contents = script_contents
        self.env_vars = env_vars or []
        self.output_var = output_var

    @transform_exception(ex.RemoteCommandException, p_ex.ImageValidationError)
    def validate(self, remote, test_only=False,
                 image_arguments=None, **kwargs):
        """Attempts to validate by running a script on the image.

        :param remote: A remote socket to the instance.
        :param test_only: If true, all validators will only verify that a
            desired state is present, and fail if it is not. If false, all
            validators will attempt to enforce the desired state if possible,
            and succeed if this enforcement succeeds.
        :param image_arguments: A dictionary of image argument values keyed by
            argument name.
            Note that the key SIV_TEST_ONLY will be set to 1 if the script
            should test_only and 0 otherwise; all scripts should act on this
            input if possible. The key SIV_DISTRO will also contain the
            distro representation, per `lsb_release -is`.
        :raises ImageValidationError: If validation fails.
        """

        arguments = copy.deepcopy(image_arguments)
        arguments[self.TEST_ONLY_KEY] = 1 if test_only else 0
        script = "\n".join(["%(env_vars)s",
                            "%(script)s"])
        env_vars = "\n".join("export %s=%s" % (key, value) for (key, value)
                             in six.iteritems(arguments)
                             if key in self.env_vars)
        script = script % {"env_vars": env_vars,
                           "script": self.script_contents.decode('utf-8')}
        path = '/tmp/%s.sh' % uuidutils.generate_uuid()
        remote.write_file_to(path, script, run_as_root=True)
        _sudo(remote, 'chmod +x %s' % path)
        code, stdout = _sudo(remote, path)
        if self.output_var:
            image_arguments[self.output_var] = stdout


class SaharaCopyScriptValidator(SaharaImageValidatorBase):
    """A validator that copy a script to the instance."""

    SPEC_SCHEMA = {
        "title": "SaharaCopyScriptValidator",
        "oneOf": [
            {
                "type": "object",
                "minProperties": 1,
                "maxProperties": 1,
                "additionalProperties": {
                    "type": "object",
                    "properties": {
                        "output": {
                            "type": "string",
                            "minLength": 1
                        },
                        "inline": {
                            "type": "string",
                            "minLength": 1
                        }
                    },
                }
            },
            {
                "type": "string"
            }
        ]
    }

    @classmethod
    def from_spec(cls, spec, validator_map, resource_roots, package='sahara'):
        """Builds a copy script validator from a specification.

        :param spec: May be a string or a single-length dictionary of name to
            configuration values. Configuration values include:
            env_vars: A list of environment variable names to send to the
                script.
            output: A key into which to put the stdout of the script in the
                image_arguments of the validation run.
        :param validator_map: A map of validator name to class.
        :param resource_roots: The roots from which relative paths to
            resources (scripts and such) will be referenced. Any resource will
            be pulled from the first path in the list at which a file exists.
        :return: A validator that will copy a script to the image.
        """
        jsonschema.validate(spec, cls.SPEC_SCHEMA)

        script_contents = None
        if isinstance(spec, six.string_types):
            script_path = spec
            output_var = None
        else:
            script_path, properties = list(six.iteritems(spec))[0]
            output_var = properties.get('output', None)
            script_contents = properties.get('inline')

        if not script_contents:
            for root in resource_roots:
                file_path = path.join(root, script_path)
                script_contents = utils.try_get_file_text(file_path, package)
                if script_contents:
                    break
        script_name = script_path.split('/')[2]

        if not script_contents:
            raise p_ex.ImageValidationSpecificationError(
                _("Script %s not found in any resource roots.") % script_path)

        return SaharaCopyScriptValidator(script_contents, script_name,
                                         output_var)

    def __init__(self, script_contents, script_name, output_var=None):
        """Constructor method.

        :param script_contents: A string representation of the script.
        :param output_var: A key into which to put the stdout of the script in
            the image_arguments of the validation run.
        :return: A SaharaScriptValidator.
        """
        self.script_contents = script_contents
        self.script_name = script_name
        self.output_var = output_var

    @transform_exception(ex.RemoteCommandException, p_ex.ImageValidationError)
    def validate(self, remote, test_only=False,
                 image_arguments=None, **kwargs):
        """Attempts to validate by running a script on the image.

        :param remote: A remote socket to the instance.
        :param test_only: If true, all validators will only verify that a
            desired state is present, and fail if it is not. If false, all
            validators will attempt to enforce the desired state if possible,
            and succeed if this enforcement succeeds.
        :param image_arguments: A dictionary of image argument values keyed by
            argument name.
            Note that the key SIV_TEST_ONLY will be set to 1 if the script
            should test_only and 0 otherwise; all scripts should act on this
            input if possible. The key SIV_DISTRO will also contain the
            distro representation, per `lsb_release -is`.
        :raises ImageValidationError: If validation fails.
        """

        arguments = copy.deepcopy(image_arguments)
        arguments[self.TEST_ONLY_KEY] = 1 if test_only else 0
        script = "\n".join(["%(script)s"])
        script = script % {"script": self.script_contents}
        path = '/tmp/%s' % self.script_name
        remote.write_file_to(path, script, run_as_root=True)


@six.add_metaclass(abc.ABCMeta)
class SaharaAggregateValidator(SaharaImageValidatorBase):
    """An abstract class representing an ordered list of other validators."""

    SPEC_SCHEMA = SaharaImageValidator.ORDERED_VALIDATORS_SCHEMA

    @classmethod
    def from_spec(cls, spec, validator_map, resource_roots, package='sahara'):
        """Builds the aggregate validator from a specification.

        :param spec: A list of validator definitions, each of which is a
            single-length dictionary of name to configuration values.
        :param validator_map: A map of validator name to class.
        :param resource_roots: The roots from which relative paths to
            resources (scripts and such) will be referenced. Any resource will
            be pulled from the first path in the list at which a file exists.
        :return: An aggregate validator.
        """
        jsonschema.validate(spec, cls.SPEC_SCHEMA)
        validators = cls.from_spec_list(spec, validator_map, resource_roots,
                                        package)
        return cls(validators)

    def __init__(self, validators):
        self.validators = validators


class SaharaAnyValidator(SaharaAggregateValidator):
    """A list of validators, only one of which must succeed."""

    def _try_all(self, remote, test_only=False,
                 image_arguments=None, **kwargs):
        results = []
        for validator in self.validators:
            result = validator.try_validate(remote, test_only=test_only,
                                            image_arguments=image_arguments,
                                            **kwargs)
            results.append(result)
            if result:
                break
        return results

    def validate(self, remote, test_only=False,
                 image_arguments=None, **kwargs):
        """Attempts to validate any of the contained validators.

        Note that if test_only=False, this validator will first run all
        contained validators using test_only=True, and succeed immediately
        should any pass validation. If all fail, it will only then run them
        using test_only=False, and again succeed immediately should any pass.

        :param remote: A remote socket to the instance.
        :param test_only: If true, all validators will only verify that a
            desired state is present, and fail if it is not. If false, all
            validators will attempt to enforce the desired state if possible,
            and succeed if this enforcement succeeds.
        :param image_arguments: A dictionary of image argument values keyed by
            argument name.
        :raises ImageValidationError: If validation fails.
        """
        results = self._try_all(remote, test_only=True,
                                image_arguments=image_arguments)
        if not test_only and not any(results):
            results = self._try_all(remote, test_only=False,
                                    image_arguments=image_arguments)
        if not any(results):
            raise p_ex.AllValidationsFailedError(result.exception for result
                                                 in results)


class SaharaAllValidator(SaharaAggregateValidator):
    """A list of validators, all of which must succeed."""

    def validate(self, remote, test_only=False, image_arguments=None,
                 **kwargs):
        """Attempts to validate all of the contained validators.

        :param remote: A remote socket to the instance.
        :param test_only: If true, all validators will only verify that a
            desired state is present, and fail if it is not. If false, all
            validators will attempt to enforce the desired state if possible,
            and succeed if this enforcement succeeds.
        :param image_arguments: A dictionary of image argument values keyed by
            argument name.
        :raises ImageValidationError: If validation fails.
        """
        for validator in self.validators:
            validator.validate(remote, test_only=test_only,
                               image_arguments=image_arguments)


class SaharaOSCaseValidator(SaharaImageValidatorBase):
    """A validator which will take different actions depending on distro."""

    _distro_tuple = collections.namedtuple('Distro', ['distro', 'validator'])

    SPEC_SCHEMA = {
        "type": "array",
        "minLength": 1,
        "items": {
            "type": "object",
            "minProperties": 1,
            "maxProperties": 1,
            "additionalProperties":
                SaharaImageValidator.ORDERED_VALIDATORS_SCHEMA,
        }
    }

    @classmethod
    def from_spec(cls, spec, validator_map, resource_roots, package='sahara'):
        """Builds an os_case validator from a specification.

        :param spec: A list of single-length dictionaries. The key of each is
            a distro or family name and the value under each key is a list of
            validators (all of which must succeed.)
        :param validator_map: A map of validator name to class.
        :param resource_roots: The roots from which relative paths to
            resources (scripts and such) will be referenced. Any resource will
            be pulled from the first path in the list at which a file exists.
        :return: A SaharaOSCaseValidator.
        """
        jsonschema.validate(spec, cls.SPEC_SCHEMA)
        distros = itertools.chain(*(six.iteritems(distro_spec)
                                    for distro_spec in spec))
        distros = [
            cls._distro_tuple(key, SaharaAllValidator.from_spec(
                value, validator_map, resource_roots, package))
            for (key, value) in distros]
        return cls(distros)

    def __init__(self, distros):
        """Constructor method.

        :param distros: A list of distro tuples (distro, list of validators).
        """
        self.distros = distros

    def validate(self, remote, test_only=False,
                 image_arguments=None, **kwargs):
        """Attempts to validate depending on distro.

        May match the OS by specific distro or by family (centos may match
        "centos" or "redhat", for instance.) If multiple keys match the
        distro, only the validators under the first matched key will be run.
        If no keys match, no validators are run, and validation proceeds.

        :param remote: A remote socket to the instance.
        :param test_only: If true, all validators will only verify that a
            desired state is present, and fail if it is not. If false, all
            validators will attempt to enforce the desired state if possible,
            and succeed if this enforcement succeeds.
        :param image_arguments: A dictionary of image argument values keyed by
            argument name.
        :raises ImageValidationError: If validation fails.
        """
        env_distro = image_arguments[self.DISTRO_KEY]
        family = self._DISTRO_FAMILES.get(env_distro)
        matches = {env_distro, family} if family else {env_distro}
        for distro, validator in self.distros:
            if distro in matches:
                validator.validate(
                    remote, test_only=test_only,
                    image_arguments=image_arguments)
                break


class SaharaArgumentCaseValidator(SaharaImageValidatorBase):
    """A validator which will take different actions depending on distro."""

    SPEC_SCHEMA = {
        "type": "object",
        "properties": {
            "argument_name": {
                "type": "string",
                "minLength": 1
            },
            "cases": {
                "type": "object",
                "minProperties": 1,
                "additionalProperties":
                    SaharaImageValidator.ORDERED_VALIDATORS_SCHEMA,
            },
        },
        "additionalProperties": False,
        "required": ["argument_name", "cases"]
    }

    @classmethod
    def from_spec(cls, spec, validator_map, resource_roots, package='sahara'):
        """Builds an argument_case validator from a specification.

        :param spec: A dictionary with two items: "argument_name", containing
            a string indicating the argument to be checked, and "cases", a
            dictionary. The key of each item in the dictionary is a value
            which may or may not match the argument value, and the value is
            a list of validators to be run in case it does.
        :param validator_map: A map of validator name to class.
        :param resource_roots: The roots from which relative paths to
            resources (scripts and such) will be referenced. Any resource will
            be pulled from the first path in the list at which a file exists.
        :return: A SaharaArgumentCaseValidator.
        """
        jsonschema.validate(spec, cls.SPEC_SCHEMA)
        argument_name = spec['argument_name']
        cases = {key: SaharaAllValidator.from_spec(
                 value, validator_map, resource_roots, package)
                 for key, value in six.iteritems(spec['cases'])}

        return cls(argument_name, cases)

    def __init__(self, argument_name, cases):
        """Constructor method.

        :param argument_name: The name of an argument.
        :param cases: A dictionary of possible argument value to a
            sub-validator to run in case of a match.
        """
        self.argument_name = argument_name
        self.cases = cases

    def validate(self, remote, test_only=False,
                 image_arguments=None, **kwargs):
        """Attempts to validate depending on argument value.

        :param remote: A remote socket to the instance.
        :param test_only: If true, all validators will only verify that a
            desired state is present, and fail if it is not. If false, all
            validators will attempt to enforce the desired state if possible,
            and succeed if this enforcement succeeds.
        :param image_arguments: A dictionary of image argument values keyed by
            argument name.
        :raises ImageValidationError: If validation fails.
        """
        arg = self.argument_name
        if arg not in image_arguments:
            raise p_ex.ImageValidationError(
                _("Argument {name} not found.").format(name=arg))
        value = image_arguments[arg]
        if value in self.cases:
            self.cases[value].validate(
                remote, test_only=test_only,
                image_arguments=image_arguments)


class SaharaArgumentSetterValidator(SaharaImageValidatorBase):
    """A validator which sets a specific argument to a specific value."""

    SPEC_SCHEMA = {
        "type": "object",
        "properties": {
            "argument_name": {
                "type": "string",
                "minLength": 1
            },
            "value": {
                "type": "string",
                "minLength": 1
            },
        },
        "additionalProperties": False,
        "required": ["argument_name", "value"]
    }

    @classmethod
    def from_spec(cls, spec, validator_map, resource_roots, package='sahara'):
        """Builds an argument_set validator from a specification.

        :param spec: A dictionary with two items: "argument_name", containing
            a string indicating the argument to be set, and "value", a value
            to which to set that argument.
        :param validator_map: A map of validator name to class.
        :param resource_roots: The roots from which relative paths to
            resources (scripts and such) will be referenced. Any resource will
            be pulled from the first path in the list at which a file exists.
        :return: A SaharaArgumentSetterValidator.
        """
        jsonschema.validate(spec, cls.SPEC_SCHEMA)
        argument_name = spec['argument_name']
        value = spec['value']

        return cls(argument_name, value)

    def __init__(self, argument_name, value):
        """Constructor method.

        :param argument_name: The name of an argument.
        :param value: A value to which to set that argument.
        """
        self.argument_name = argument_name
        self.value = value

    def validate(self, remote, test_only=False,
                 image_arguments=None, **kwargs):
        """Attempts to validate depending on argument value.

        :param remote: A remote socket to the instance.
        :param test_only: If true, all validators will only verify that a
            desired state is present, and fail if it is not. If false, all
            validators will attempt to enforce the desired state if possible,
            and succeed if this enforcement succeeds.
        :param image_arguments: A dictionary of image argument values keyed by
            argument name.
        """
        image_arguments[self.argument_name] = self.value


def _sudo(remote, cmd, **kwargs):
    return remote.execute_command(cmd, run_as_root=True, **kwargs)



# Copyright (c) 2016 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections

from oslo_config import cfg
from oslo_log import log as logging

from sahara import conductor
from sahara import context
from sahara import exceptions
from sahara.i18n import _
from sahara.plugins import health_check_base
from sahara.service.health import common
from sahara.utils import cluster as cluster_utils


cond = conductor.API
CONF = cfg.CONF
LOG = logging.getLogger(__name__)


class CannotVerifyError(exceptions.SaharaException):
    code = "CANNOT_VERIFY"
    message_template = _("Cannot verify cluster. Reason: %s")

    def __init__(self, reason):
        message = self.message_template % reason
        super(CannotVerifyError, self).__init__(message)


def get_possible_ops():
    return common.VERIFICATIONS_OPS


def verification_exists(cluster):
    try:
        if cluster.verification is not None:
            return True
    except (AttributeError, KeyError):
        return False


def validate_verification_ops(cluster, data):
    status = data.get('verification', {}).get('status', None)
    if not status:
        # update doesn't affect verifications
        return False
    if len(list(data.keys())) != 1:
        raise CannotVerifyError(
            _("Can't update verification with other updates"))
    if status == common.VERIFICATIONS_START_OPS:
        return validate_verification_start(cluster)


def clean_verification_data(cluster):
    cluster = cond.cluster_get(context.ctx(), cluster)
    if verification_exists(cluster):
        try:
            vid = cluster.verification.id
            cond.cluster_verification_delete(context.ctx(), vid)
        except exceptions.NotFoundException:
            LOG.debug("Verification data already cleaned")


def validate_verification_start(cluster):
    if not CONF.cluster_verifications.verification_enable:
        raise CannotVerifyError(_("All verifications are disabled"))
    ctx = context.ctx()
    cluster = cond.cluster_get(ctx, cluster)
    if not cluster or cluster.status != cluster_utils.CLUSTER_STATUS_ACTIVE:
        raise CannotVerifyError(_("Cluster is not active or doesn't exists"))
    if not verification_exists(cluster):
        return True
    if cluster.verification.status == common.HEALTH_STATUS_CHECKING:
        raise CannotVerifyError(
            _("Cluster verification in state %s")
            % common.HEALTH_STATUS_CHECKING)
    return True


def _prepare_verification_running(ctx, cluster):
    if verification_exists(cluster):
        vid = cluster.verification.id
        # to delete all data related to the previous one
        cond.cluster_verification_delete(ctx, vid)
    return (cond.cluster_verification_add(
        ctx, cluster, {'status': common.HEALTH_STATUS_CHECKING}),
        cond.cluster_get(ctx, cluster))


def _execute_health_checks(ctx, cluster):
    health_checks = health_check_base.get_health_checks(cluster)
    actual = []
    with context.ThreadGroup() as tg:
        for check in health_checks:
            actual_check = check(cluster)
            actual.append(actual_check)
            tg.spawn('health-check-exc', actual_check.execute)


def _decide_status_for_verification(ctx, verification):
    ver = cond.cluster_verification_get(ctx, verification)
    cnt = collections.Counter()
    for check in ver.checks:
        cnt[check.status] += 1

    if cnt[common.HEALTH_STATUS_GREEN] == len(ver.checks):
        decided_status = common.HEALTH_STATUS_GREEN
    elif cnt[common.HEALTH_STATUS_RED] > 0:
        decided_status = common.HEALTH_STATUS_RED
    else:
        decided_status = common.HEALTH_STATUS_YELLOW

    return cond.cluster_verification_update(
        context.ctx(), ver.id, {'status': decided_status})


def verification_run(cluster):
    ctx = context.ctx()
    LOG.debug("Running verification for the cluster")
    ver, cluster = _prepare_verification_running(ctx, cluster)
    _execute_health_checks(ctx, cluster)
    return _decide_status_for_verification(ctx, ver)


def handle_verification(cluster, values):
    cluster = cond.cluster_get(context.ctx(), cluster)
    context.set_current_cluster_id(cluster.id)
    values = {} if not values else values
    status = values.get('verification', {}).get('status', None)
    if status == common.VERIFICATIONS_START_OPS:
        verification_run(cluster)


def update_verification_required(values):
    if values.get('verification', {}).get('status', None):
        return True
    return False


def get_verification_periodic_interval():
    return (CONF.cluster_verifications.verification_periodic_interval if
            CONF.cluster_verifications.verification_enable else -1)



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import six

from sahara import exceptions as ex
from sahara.i18n import _


def check_tenant_for_delete(context, object):
    if object.tenant_id != context.tenant_id:
        raise ex.DeletionFailed(
            _("{object} with id '{id}' could not be deleted because "
              "it wasn't created in this tenant").format(
                object=type(object).__name__, id=object.id))


def check_tenant_for_update(context, object):
    if object.tenant_id != context.tenant_id:
        raise ex.UpdateFailedException(
            object.id,
            _("{object} with id '%s' could not be updated because "
              "it wasn't created in this tenant").format(
                object=type(object).__name__))


def check_protected_from_delete(object):
    if object.is_protected:
        raise ex.DeletionFailed(
            _("{object} with id '{id}' could not be deleted because "
              "it's marked as protected").format(
                object=type(object).__name__, id=object.id))


def check_protected_from_update(object, data):
    if object.is_protected and data.get('is_protected', True):
        # Okay, the only thing we can allow here is a change
        # to 'is_public', so we have to make sure no other values
        # are changing
        if 'is_public' in data:
            obj = object.to_dict()
            if all(k == 'is_public' or (
                    k in obj and obj[k] == v) for k, v in six.iteritems(data)):
                return

        raise ex.UpdateFailedException(
            object.id,
            _("{object} with id '%s' could not be updated "
              "because it's marked as protected").format(
                object=type(object).__name__))



# Copyright (c) 2017 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import mock

from oslo_utils import uuidutils
import testtools

from sahara import exceptions as ex
from sahara.service.edp.job_binaries.internal_db import implementation


class TestInternalDBType(testtools.TestCase):

    def setUp(self):
        super(TestInternalDBType, self).setUp()
        self.internal_db = implementation.InternalDBType()

    def test_validate_job_location_format(self):

        self.assertFalse(self.internal_db.
                         validate_job_location_format(''))
        self.assertFalse(self.internal_db.
                         validate_job_location_format('invalid-scheme://'))
        self.assertFalse(self.internal_db.
                         validate_job_location_format('internal-db://abc'))
        self.assertTrue(self.internal_db.
                        validate_job_location_format(
                            'internal-db://' + uuidutils.generate_uuid()))

    @mock.patch('sahara.conductor.API.job_binary_internal_get_raw_data')
    def test_copy_binary_to_cluster(self, conductor_get_raw_data):
        remote = mock.Mock()
        context = mock.Mock()
        conductor_get_raw_data.return_value = 'ok'
        job_binary = mock.Mock()
        job_binary.name = 'test'
        job_binary.url = 'internal-db://somebinary'

        res = self.internal_db.copy_binary_to_cluster(job_binary,
                                                      context=context,
                                                      remote=remote)

        self.assertEqual('/tmp/test', res)
        remote.write_file_to.assert_called_with(
            '/tmp/test',
            'ok')

    @mock.patch('sahara.conductor.API.job_binary_internal_get_raw_data')
    def test_get_raw_data(self, conductor_get_raw_data):
        context = mock.Mock()
        conductor_get_raw_data.return_value = 'ok'
        job_binary = mock.Mock()
        job_binary.url = 'internal-db://somebinary'

        self.internal_db.get_raw_data(job_binary,
                                      context=context)

    @mock.patch('sahara.service.validations.edp.base.'
                'check_job_binary_internal_exists')
    def test_data_validation(self, check_exists):
        data = {
            'url': '',
            'description': 'empty url'
        }
        with testtools.ExpectedException(ex.InvalidDataException):
            self.internal_db.validate(data)

        data = {
            'url': 'invalid-url://',
            'description': 'not empty, but invalid url'
        }
        with testtools.ExpectedException(ex.InvalidDataException):
            self.internal_db.validate(data)

        data = {
            'url': 'internal-db://must-be-uuid',
            'description': 'correct scheme, but not netloc is not uuid'
        }
        with testtools.ExpectedException(ex.InvalidDataException):
            self.internal_db.validate(data)

        data = {
            'url': 'internal-db://' + uuidutils.generate_uuid(),
            'description': 'correct scheme and netloc'
        }
        self.internal_db.validate(data)



# Copyright (c) 2016 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import mock

import sys
guestfs = mock.Mock()
sys.modules['guestfs'] = guestfs

from sahara.cli.image_pack import api
from sahara.tests.unit import base


class TestSaharaImagePackAPI(base.SaharaTestCase):

    def setUp(self):
        super(TestSaharaImagePackAPI, self).setUp()

    def tearDown(self):
        super(TestSaharaImagePackAPI, self).tearDown()

    @mock.patch('sahara.cli.image_pack.api.guestfs')
    @mock.patch('sahara.cli.image_pack.api.plugins_base')
    @mock.patch('sahara.cli.image_pack.api.LOG')
    def test_pack_image_call(self, mock_log, mock_plugins_base, mock_guestfs):
        guest = mock.Mock()
        mock_guestfs.GuestFS = mock.Mock(return_value=guest)
        guest.inspect_os = mock.Mock(return_value=['/dev/something1'])
        plugin = mock.Mock()
        mock_plugins_base.PLUGINS = mock.Mock(
            get_plugin=mock.Mock(return_value=plugin))

        api.pack_image(
            "image_path", "plugin_name", "plugin_version",
            {"anarg": "avalue"}, root_drive=None, test_only=False)

        guest.add_drive_opts.assert_called_with("image_path", format="qcow2")
        guest.set_network.assert_called_with(True)
        guest.launch.assert_called_once_with()
        guest.mount.assert_called_with('/dev/something1', '/')
        guest.sh.assert_called_with("echo Testing sudo without tty...")
        guest.sync.assert_called_once_with()
        guest.umount_all.assert_called_once_with()
        guest.close.assert_called_once_with()

    @mock.patch('sahara.cli.image_pack.api.plugins_base')
    def test_get_plugin_arguments(self, mock_plugins_base):
        api.setup_plugins()
        mock_plugins_base.setup_plugins.assert_called_once_with()
        mock_PLUGINS = mock.Mock()
        mock_plugins_base.PLUGINS = mock_PLUGINS
        mock_plugin = mock.Mock()
        mock_plugin.get_versions = mock.Mock(return_value=['1'])
        mock_plugin.get_image_arguments = mock.Mock(
            return_value=["Argument!"])
        mock_PLUGINS.get_plugin = mock.Mock(return_value=mock_plugin)
        result = api.get_plugin_arguments('Plugin!')
        mock_plugin.get_versions.assert_called_once_with()
        mock_plugin.get_image_arguments.assert_called_once_with('1')
        self.assertEqual(result, {'1': ['Argument!']})



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re

import jsonschema
from oslo_utils import uuidutils
import six

from sahara.service.edp.job_binaries import manager as jb_manager


@jsonschema.FormatChecker.cls_checks('valid_name_hostname')
def validate_name_hostname_format(entry):
    if not isinstance(entry, six.string_types) or not entry:
        # should fail type or length validation
        return True

    res = re.match(r"^(([a-zA-Z]|[a-zA-Z][a-zA-Z0-9\-]"
                   r"*[a-zA-Z0-9])\.)*([A-Za-z]|[A-Za-z]"
                   r"[A-Za-z0-9\-]*[A-Za-z0-9])$", entry)
    return res is not None


@jsonschema.FormatChecker.cls_checks('valid_name')
def validate_name_format(entry):
    if not isinstance(entry, six.string_types):
        # should fail type validation
        return True

    res = re.match(r"^[a-zA-Z0-9][a-zA-Z0-9\-_\.]*$", entry)
    return res is not None


@jsonschema.FormatChecker.cls_checks('valid_keypair_name')
def validate_keypair_name_format(entry):
    if not isinstance(entry, six.string_types):
        # should fail type validation
        return True

    # this follows the validation put forth by nova for keypair names
    res = re.match(r'^[a-zA-Z0-9\-_ ]+$', entry)
    return res is not None


@jsonschema.FormatChecker.cls_checks('valid_job_location')
def validate_job_location_format(entry):
    if not isinstance(entry, six.string_types):
        # should fail type validation
        return True

    return jb_manager.JOB_BINARIES \
                     .get_job_binary_by_url(entry) \
                     .validate_job_location_format(entry)


@jsonschema.FormatChecker.cls_checks('valid_tag')
def validate_valid_tag_format(entry):
    if not isinstance(entry, six.string_types):
        # should fail type validation
        return True

    res = re.match(r"^(([a-zA-Z0-9]|[a-zA-Z0-9][a-zA-Z0-9\-_]"
                   r"*[a-zA-Z0-9])\.)*([A-Za-z0-9]|[A-Za-z0-9]"
                   r"[A-Za-z0-9\-_]*[A-Za-z0-9])$", entry)
    return res is not None


@jsonschema.FormatChecker.cls_checks('uuid')
def validate_uuid_format(entry):
    if not isinstance(entry, six.string_types):
        # should fail type validation
        return True

    return uuidutils.is_uuid_like(entry)


@jsonschema.FormatChecker.cls_checks('posix_path')
def validate_posix_path(entry):
    if not isinstance(entry, six.string_types):
        # should fail type validation
        return True

    res = re.match("^(/([A-Z]|[a-z]|[0-9]|\-|_)+)+$", entry)
    return res is not None


def is_config(cls, instance):
    # configs should be dict
    if not isinstance(instance, dict):
        return False

    # check dict content
    for applicable_target, configs in six.iteritems(instance):
        # upper-level dict keys (applicable targets) should be strings
        if not isinstance(applicable_target, six.string_types):
            return False

        # upper-level dict values should be dicts
        if not isinstance(configs, dict):
            return False

        # check internal dict content
        for config_name, config_value in six.iteritems(configs):
            # internal dict keys should be strings
            if not isinstance(config_name, six.string_types):
                return False

            # internal dict values should be strings or integers or bools
            if not isinstance(config_value,
                              (six.string_types, six.integer_types)):
                return False

    return True


def is_simple_config(cls, instance):
    # configs should be dict
    if not isinstance(instance, dict):
        return False

    # check dict content
    for conf_name, conf_value in six.iteritems(instance):
        # keys should be strings, values should be int, string or bool
        if not isinstance(conf_name, six.string_types):
            return False
        if not isinstance(conf_value,
                          (six.string_types, six.integer_types)):
            return False
    return True


def is_flavor_type(cls, instance):
    try:
        int(instance)
    except (ValueError, TypeError):
        return (isinstance(instance, six.string_types)
                and uuidutils.is_uuid_like(instance))
    return (isinstance(instance, six.integer_types + six.string_types)
            and type(instance) != bool)


class ApiValidator(jsonschema.Draft4Validator):

    TYPE_CHECKER = jsonschema.Draft4Validator.TYPE_CHECKER.redefine_many({
        "configs": is_config,
        "flavor": is_flavor_type,
        "simple_config": is_simple_config,
    })

    def __init__(self, schema, resolver=None,
                 format_checker=jsonschema.FormatChecker()):
        super(ApiValidator, self).__init__(
            schema, format_checker=format_checker)



# Copyright (c) 2015 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import mock

import six

from sahara import conductor as cond
from sahara import context
from sahara.plugins import recommendations_utils as ru
from sahara.tests.unit import base as b

conductor = cond.API


class Configs(object):
    def __init__(self, configs):
        self.configs = configs

    def to_dict(self):
        return self.configs


class FakeObject(object):
    def __init__(self, **kwargs):
        for attr in six.iterkeys(kwargs):
            setattr(self, attr, kwargs.get(attr))


class TestProvidingRecommendations(b.SaharaWithDbTestCase):
    @mock.patch('sahara.utils.openstack.nova.get_flavor')
    def test_get_recommended_node_configs_medium_flavor(
            self, fake_flavor):
        ng = FakeObject(flavor_id="fake_flavor", node_configs=Configs({}))
        cl = FakeObject(cluster_configs=Configs({}))
        fake_flavor.return_value = FakeObject(ram=4096, vcpus=2)
        observed = ru.HadoopAutoConfigsProvider(
            {}, [], cl, False)._get_recommended_node_configs(ng)
        self.assertEqual({
            'mapreduce.reduce.memory.mb': 768,
            'mapreduce.map.java.opts': '-Xmx307m',
            'mapreduce.map.memory.mb': 384,
            'mapreduce.reduce.java.opts': '-Xmx614m',
            'yarn.app.mapreduce.am.resource.mb': 384,
            'yarn.app.mapreduce.am.command-opts': '-Xmx307m',
            'mapreduce.task.io.sort.mb': 153,
            'yarn.nodemanager.resource.memory-mb': 3072,
            'yarn.scheduler.minimum-allocation-mb': 384,
            'yarn.scheduler.maximum-allocation-mb': 3072,
            'yarn.nodemanager.vmem-check-enabled': 'false'
        }, observed)

    @mock.patch('sahara.utils.openstack.nova.get_flavor')
    def test_get_recommended_node_configs_small_flavor(
            self, fake_flavor):
        ng = FakeObject(flavor_id="fake_flavor", node_configs=Configs({}))
        cl = FakeObject(cluster_configs=Configs({}))
        fake_flavor.return_value = FakeObject(ram=2048, vcpus=1)
        observed = ru.HadoopAutoConfigsProvider(
            {'node_configs': {}, 'cluster_configs': {}}, [], cl, False,
        )._get_recommended_node_configs(ng)
        self.assertEqual({
            'mapreduce.reduce.java.opts': '-Xmx409m',
            'yarn.app.mapreduce.am.resource.mb': 256,
            'mapreduce.reduce.memory.mb': 512,
            'mapreduce.map.java.opts': '-Xmx204m',
            'yarn.app.mapreduce.am.command-opts': '-Xmx204m',
            'mapreduce.task.io.sort.mb': 102,
            'mapreduce.map.memory.mb': 256,
            'yarn.nodemanager.resource.memory-mb': 2048,
            'yarn.scheduler.minimum-allocation-mb': 256,
            'yarn.nodemanager.vmem-check-enabled': 'false',
            'yarn.scheduler.maximum-allocation-mb': 2048,
        }, observed)

    def test_merge_configs(self):
        provider = ru.HadoopAutoConfigsProvider({}, None, None, False)
        initial_configs = {
            'cat': {
                'talk': 'meow',
            },
            'bond': {
                'name': 'james'
            }
        }

        extra_configs = {
            'dog': {
                'talk': 'woof'
            },
            'bond': {
                'extra_name': 'james bond'
            }
        }

        expected = {
            'cat': {
                'talk': 'meow',
            },
            'dog': {
                'talk': 'woof'
            },
            'bond': {
                'name': 'james',
                'extra_name': 'james bond'
            }
        }
        self.assertEqual(
            expected, provider._merge_configs(initial_configs, extra_configs))

    @mock.patch('sahara.utils.openstack.nova.get_flavor')
    @mock.patch('sahara.plugins.recommendations_utils.conductor.'
                'node_group_update')
    @mock.patch('sahara.plugins.recommendations_utils.conductor.'
                'cluster_update')
    def test_apply_recommended_configs(self, cond_cluster, cond_node_group,
                                       fake_flavor):
        class TestProvider(ru.HadoopAutoConfigsProvider):
            def get_datanode_name(self):
                return "dog_datanode"
        fake_flavor.return_value = FakeObject(ram=2048, vcpus=1)
        to_tune = {
            'cluster_configs': {
                'dfs.replication': ('dfs', 'replica')
            },
            'node_configs': {
                'mapreduce.task.io.sort.mb': ('bond', 'extra_name')
            }
        }

        fake_plugin_configs = [
            FakeObject(applicable_target='dfs', name='replica',
                       default_value=3)]
        fake_ng = FakeObject(
            use_autoconfig=True,
            count=2,
            node_processes=['dog_datanode'],
            flavor_id='fake_id',
            node_configs=Configs({
                'bond': {
                    'name': 'james'
                }
            })
        )
        fake_cluster = FakeObject(
            cluster_configs=Configs({
                'cat': {
                    'talk': 'meow',
                }
            }),
            node_groups=[fake_ng],
            use_autoconfig=True,
            extra=Configs({})
        )
        v = TestProvider(
            to_tune, fake_plugin_configs, fake_cluster, False)

        v.apply_recommended_configs()
        self.assertEqual([mock.call(context.ctx(), fake_cluster, {
            'cluster_configs': {
                'cat': {
                    'talk': 'meow'
                },
                'dfs': {
                    'replica': 2
                }
            }
        }), mock.call(
            context.ctx(), fake_cluster,
            {'extra': {'auto-configured': True}})],
            cond_cluster.call_args_list)
        self.assertEqual([mock.call(context.ctx(), fake_ng, {
            'node_configs': {
                'bond': {
                    'name': 'james',
                    'extra_name': 102
                }
            }
        })], cond_node_group.call_args_list)

    @mock.patch('sahara.utils.openstack.nova.get_flavor')
    @mock.patch('sahara.plugins.recommendations_utils.conductor.'
                'node_group_update')
    @mock.patch('sahara.plugins.recommendations_utils.conductor.'
                'cluster_update')
    def test_apply_recommended_configs_no_updates(
            self, cond_cluster, cond_node_group, fake_flavor):
        fake_flavor.return_value = FakeObject(ram=2048, vcpus=1)
        to_tune = {
            'cluster_configs': {
                'dfs.replication': ('dfs', 'replica')
            },
            'node_configs': {
                'mapreduce.task.io.sort.mb': ('bond', 'extra_name')
            }
        }

        fake_plugin_configs = [
            FakeObject(applicable_target='dfs', name='replica',
                       default_value=3)]
        fake_ng = FakeObject(
            use_autoconfig=True,
            count=2,
            node_processes=['dog_datanode'],
            flavor_id='fake_id',
            node_configs=Configs({
                'bond': {
                    'extra_name': 'james bond'
                }
            })
        )
        fake_cluster = FakeObject(
            cluster_configs=Configs({
                'dfs': {
                    'replica': 1
                }
            }),
            node_groups=[fake_ng],
            use_autoconfig=True,
            extra=Configs({})
        )
        v = ru.HadoopAutoConfigsProvider(
            to_tune, fake_plugin_configs, fake_cluster, False)
        v.apply_recommended_configs()
        self.assertEqual(0, cond_node_group.call_count)
        self.assertEqual(
            [mock.call(context.ctx(), fake_cluster,
                       {'extra': {'auto-configured': True}})],
            cond_cluster.call_args_list)

    def test_correct_use_autoconfig_value(self):
        ctx = context.ctx()
        ngt1 = conductor.node_group_template_create(ctx, {
            'name': 'ngt1',
            'flavor_id': '1',
            'plugin_name': 'vanilla',
            'hadoop_version': '1'
        })
        ngt2 = conductor.node_group_template_create(ctx, {
            'name': 'ngt2',
            'flavor_id': '2',
            'plugin_name': 'vanilla',
            'hadoop_version': '1',
            'use_autoconfig': False
        })
        self.assertTrue(ngt1.use_autoconfig)
        self.assertFalse(ngt2.use_autoconfig)
        clt = conductor.cluster_template_create(ctx, {
            'name': "clt1",
            'plugin_name': 'vanilla',
            'hadoop_version': '1',
            'node_groups': [
                {
                    'count': 3,
                    "node_group_template_id": ngt1.id
                },
                {
                    'count': 1,
                    'node_group_template_id': ngt2.id
                }
            ],
            'use_autoconfig': False
        })
        cluster = conductor.cluster_create(ctx, {
            'name': 'stupid',
            'cluster_template_id': clt.id
        })
        self.assertFalse(cluster.use_autoconfig)
        for ng in cluster.node_groups:
            if ng.name == 'ngt1':
                self.assertTrue(ng.use_autoconfig)
            else:
                self.assertFalse(ng.use_autoconfig)

    @mock.patch('sahara.plugins.recommendations_utils.conductor.'
                'cluster_update')
    def test_not_autonconfigured(self, cluster_update):
        fake_cluster = FakeObject(extra=Configs({}))
        v = ru.HadoopAutoConfigsProvider({}, [], fake_cluster, True)
        v.apply_recommended_configs()
        self.assertEqual(0, cluster_update.call_count)



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_serialization import jsonutils
import sqlalchemy as sa
from sqlalchemy.dialects import mysql
from sqlalchemy.ext import mutable


class JsonEncoded(sa.TypeDecorator):
    """Represents an immutable structure as a json-encoded string."""

    impl = sa.Text

    def process_bind_param(self, value, dialect):
        if value is not None:
            value = jsonutils.dumps(value)
        return value

    def process_result_value(self, value, dialect):
        if value is not None:
            value = jsonutils.loads(value)
        return value


class MutableDict(mutable.Mutable, dict):
    @classmethod
    def coerce(cls, key, value):
        """Convert plain dictionaries to MutableDict."""
        if not isinstance(value, MutableDict):
            if isinstance(value, dict):
                return MutableDict(value)

            # this call will raise ValueError
            return mutable.Mutable.coerce(key, value)
        else:
            return value

    def update(self, e=None, **f):
        """Detect dictionary update events and emit change events."""
        dict.update(self, e, **f)
        self.changed()

    def __setitem__(self, key, value):
        """Detect dictionary set events and emit change events."""
        dict.__setitem__(self, key, value)
        self.changed()

    def __delitem__(self, key):
        """Detect dictionary del events and emit change events."""
        dict.__delitem__(self, key)
        self.changed()


class MutableList(mutable.Mutable, list):
    @classmethod
    def coerce(cls, key, value):
        """Convert plain lists to MutableList."""
        if not isinstance(value, MutableList):
            if isinstance(value, list):
                return MutableList(value)

            # this call will raise ValueError
            return mutable.Mutable.coerce(key, value)
        else:
            return value

    def append(self, value):
        """Detect list add events and emit change events."""
        list.append(self, value)
        self.changed()

    def remove(self, value):
        """Removes an item by value and emit change events."""
        list.remove(self, value)
        self.changed()

    def __setitem__(self, key, value):
        """Detect list set events and emit change events."""
        list.__setitem__(self, key, value)
        self.changed()

    def __delitem__(self, i):
        """Detect list del events and emit change events."""
        list.__delitem__(self, i)
        self.changed()


def JsonDictType():
    """Returns an SQLAlchemy Column Type suitable to store a Json dict."""
    return MutableDict.as_mutable(JsonEncoded)


def JsonListType():
    """Returns an SQLAlchemy Column Type suitable to store a Json array."""
    return MutableList.as_mutable(JsonEncoded)


def LargeBinary():
    return sa.LargeBinary().with_variant(mysql.LONGBLOB(), 'mysql')


def LongText():
    return sa.Text().with_variant(mysql.LONGTEXT(), 'mysql')



# Copyright (c) 2014 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc

import six

from sahara import conductor as c

conductor = c.API


def optional(fun):
    fun.__not_implemented__ = True
    return fun


@six.add_metaclass(abc.ABCMeta)
class JobEngine(object):
    @abc.abstractmethod
    def cancel_job(self, job_execution):
        pass

    @abc.abstractmethod
    def get_job_status(self, job_execution):
        pass

    @abc.abstractmethod
    def run_job(self, job_execution):
        pass

    @abc.abstractmethod
    def run_scheduled_job(self, job_execution):
        pass

    @abc.abstractmethod
    def validate_job_execution(self, cluster, job, data):
        pass

    @staticmethod
    @abc.abstractmethod
    def get_possible_job_config(job_type):
        return None

    @staticmethod
    @abc.abstractmethod
    def get_supported_job_types():
        return None

    @optional
    def suspend_job(self, job_execution):
        pass

    def does_engine_implement(self, fun_name):
        fun = getattr(self, fun_name)
        if not (fun and callable(fun)):
            return False
        return not hasattr(fun, '__not_implemented__')



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
import traceback

import flask
import microversion_parse
from oslo_log import log as logging
from oslo_middleware import request_id as oslo_req_id
import six
from werkzeug import datastructures

from sahara.api import microversion as mv
from sahara import context
from sahara import exceptions as ex
from sahara.i18n import _
from sahara.utils import types
from sahara.utils import wsgi


LOG = logging.getLogger(__name__)


class Rest(flask.Blueprint):
    def get(self, rule, status_code=200):
        return self._mroute('GET', rule, status_code)

    def post(self, rule, status_code=202):
        return self._mroute('POST', rule, status_code)

    def post_file(self, rule, status_code=202):
        return self._mroute('POST', rule, status_code, file_upload=True)

    def put(self, rule, status_code=202):
        return self._mroute('PUT', rule, status_code)

    def put_file(self, rule, status_code=202):
        return self._mroute('PUT', rule, status_code, file_upload=True)

    def delete(self, rule, status_code=204):
        return self._mroute('DELETE', rule, status_code)

    def patch(self, rule, status_code=202):
        return self._mroute('PATCH', rule, status_code)

    def _mroute(self, methods, rule, status_code=None, **kw):
        if isinstance(methods, six.string_types):
            methods = [methods]
        return self.route(rule, methods=methods, status_code=status_code, **kw)

    def route(self, rule, **options):
        status = options.pop('status_code', None)
        file_upload = options.pop('file_upload', False)

        def decorator(func):
            endpoint = options.pop('endpoint', func.__name__)

            def handler(**kwargs):
                context.set_ctx(None)

                LOG.debug("Rest.route.decorator.handler, kwargs={kwargs}"
                          .format(kwargs=kwargs))

                _init_resp_type(file_upload)

                # update status code
                if status:
                    flask.request.status_code = status

                kwargs.pop("tenant_id", None)
                req_id = flask.request.environ.get(oslo_req_id.ENV_REQUEST_ID)
                auth_plugin = flask.request.environ.get('keystone.token_auth')
                ctx = context.Context(
                    flask.request.headers['X-User-Id'],
                    flask.request.headers['X-Tenant-Id'],
                    flask.request.headers['X-Auth-Token'],
                    flask.request.headers['X-Service-Catalog'],
                    flask.request.headers['X-User-Name'],
                    flask.request.headers['X-Tenant-Name'],
                    flask.request.headers['X-Roles'].split(','),
                    auth_plugin=auth_plugin,
                    request_id=req_id)
                context.set_ctx(ctx)

                try:
                    if flask.request.method in ['POST', 'PUT', 'PATCH']:
                        kwargs['data'] = request_data()
                    return func(**kwargs)
                except ex.Forbidden as e:
                    return access_denied(e)
                except ex.SaharaException as e:
                    return bad_request(e)
                except Exception as e:
                    return internal_error(500, 'Internal Server Error', e)

            f_rule = "/<tenant_id>" + rule
            self.add_url_rule(rule, endpoint, handler, **options)
            self.add_url_rule(rule + '.json', endpoint, handler, **options)
            self.add_url_rule(f_rule, endpoint, handler, **options)
            self.add_url_rule(f_rule + '.json', endpoint, handler, **options)

            return func

        return decorator


def check_microversion_header():
    requested_version = get_requested_microversion()
    if not re.match(mv.VERSION_STRING_REGEX, requested_version):
        bad_request_microversion(requested_version)
    if requested_version not in mv.API_VERSIONS:
        not_acceptable_microversion(requested_version)


def add_vary_header(response):
    response.headers[mv.VARY_HEADER] = mv.OPENSTACK_API_VERSION_HEADER
    response.headers[mv.OPENSTACK_API_VERSION_HEADER] = "{} {}".format(
        mv.SAHARA_SERVICE_TYPE, get_requested_microversion())
    return response


class RestV2(Rest):
    def __init__(self, *args, **kwargs):
        super(RestV2, self).__init__(*args, **kwargs)
        self.before_request(check_microversion_header)
        self.after_request(add_vary_header)

    def route(self, rule, **options):
        status = options.pop('status_code', None)
        file_upload = options.pop('file_upload', False)

        def decorator(func):
            endpoint = options.pop('endpoint', func.__name__)

            def handler(**kwargs):
                context.set_ctx(None)

                LOG.debug("Rest.route.decorator.handler, kwargs={kwargs}"
                          .format(kwargs=kwargs))

                _init_resp_type(file_upload)

                # update status code
                if status:
                    flask.request.status_code = status

                kwargs.pop("tenant_id", None)
                req_id = flask.request.environ.get(oslo_req_id.ENV_REQUEST_ID)
                auth_plugin = flask.request.environ.get('keystone.token_auth')
                ctx = context.Context(
                    flask.request.headers['X-User-Id'],
                    flask.request.headers['X-Tenant-Id'],
                    flask.request.headers['X-Auth-Token'],
                    flask.request.headers['X-Service-Catalog'],
                    flask.request.headers['X-User-Name'],
                    flask.request.headers['X-Tenant-Name'],
                    flask.request.headers['X-Roles'].split(','),
                    auth_plugin=auth_plugin,
                    request_id=req_id)
                context.set_ctx(ctx)

                try:
                    if flask.request.method in ['POST', 'PUT', 'PATCH']:
                        kwargs['data'] = request_data()
                    return func(**kwargs)
                except ex.Forbidden as e:
                    return access_denied(e)
                except ex.SaharaException as e:
                    return bad_request(e)
                except Exception as e:
                    return internal_error(500, 'Internal Server Error', e)

            f_rule = "/<tenant_id>" + rule
            self.add_url_rule(rule, endpoint, handler, **options)
            self.add_url_rule(rule + '.json', endpoint, handler, **options)
            self.add_url_rule(f_rule, endpoint, handler, **options)
            self.add_url_rule(f_rule + '.json', endpoint, handler, **options)

            return func

        return decorator


RT_JSON = datastructures.MIMEAccept([("application/json", 1)])


def _init_resp_type(file_upload):
    """Extracts response content type."""

    # get content type from Accept header
    resp_type = flask.request.accept_mimetypes

    # url /foo.json
    if flask.request.path.endswith('.json'):
        resp_type = RT_JSON

    flask.request.resp_type = resp_type

    # set file upload flag
    flask.request.file_upload = file_upload


def render(res=None, resp_type=None, status=None, name=None, **kwargs):
    if not res and type(res) is not types.Page:
        res = {}
    if type(res) is dict:
        res.update(kwargs)

    elif type(res) is types.Page:

        result = {name: [item.to_dict() for item in res]}
        result.update(kwargs)
        if res.prev or res.next or ('marker' in get_request_args()):
            result["markers"] = {"prev": res.prev, "next": res.next}
        res = result
    elif kwargs:
        # can't merge kwargs into the non-dict res
        abort_and_log(500,
                      _("Non-dict and non-empty kwargs passed to render"))

    status_code = getattr(flask.request, 'status_code', None)
    if status:
        status_code = status
    if not status_code:
        status_code = 200

    if not resp_type:
        resp_type = getattr(flask.request, 'resp_type', RT_JSON)

    if not resp_type:
        resp_type = RT_JSON

    serializer = None
    if "application/json" in resp_type:
        resp_type = RT_JSON
        serializer = wsgi.JSONDictSerializer()
    else:
        raise ex.InvalidDataException(
            _("Content type '%s' isn't supported") % resp_type)

    body = serializer.serialize(res)
    resp_type = str(resp_type)

    return flask.Response(response=body, status=status_code,
                          mimetype=resp_type)


def request_data():
    if hasattr(flask.request, 'parsed_data'):
        return flask.request.parsed_data

    if (flask.request.content_length is None
            or not flask.request.content_length > 0):
        LOG.debug("Empty body provided in request")
        return dict()

    if flask.request.file_upload:
        return flask.request.data

    deserializer = None
    content_type = flask.request.mimetype
    if not content_type or content_type in RT_JSON:
        deserializer = wsgi.JSONDeserializer()
    else:
        raise ex.InvalidDataException(
            _("Content type '%s' isn't supported") % content_type)

    # parsed request data to avoid unwanted re-parsings
    parsed_data = deserializer.deserialize(flask.request.data)['body']
    flask.request.parsed_data = parsed_data

    return flask.request.parsed_data


def get_request_args():
    return flask.request.args


def get_requested_microversion():
    requested_version = microversion_parse.get_version(
        flask.request.headers,
        mv.SAHARA_SERVICE_TYPE
    )
    if requested_version is None:
        requested_version = mv.MIN_API_VERSION
    elif requested_version == mv.LATEST:
        requested_version = mv.MAX_API_VERSION
    return requested_version


def abort_and_log(status_code, descr, exc=None):
    LOG.error("Request aborted with status code {code} and "
              "message '{message}'".format(code=status_code, message=descr))

    if exc is not None:
        LOG.error(traceback.format_exc())

    flask.abort(status_code, description=descr)


def render_error_message(error_code, error_message, error_name, **msg_kwargs):
    message = {
        "error_code": error_code,
        "error_message": error_message,
        "error_name": error_name
    }

    message.update(**msg_kwargs)

    resp = render(message)
    resp.status_code = error_code

    return resp


def not_acceptable_microversion(requested_version):
    message = ("Version {} is not supported by the API. "
               "Minimum is {} and maximum is {}.".format(
                   requested_version,
                   mv.MIN_API_VERSION,
                   mv.MAX_API_VERSION
               ))
    resp = render_error_message(
        mv.NOT_ACCEPTABLE_STATUS_CODE,
        message,
        mv.NOT_ACCEPTABLE_STATUS_NAME,
        max_version=mv.MAX_API_VERSION,
        min_version=mv.MIN_API_VERSION
    )
    flask.abort(resp)


def bad_request_microversion(requested_version):
    message = ("API Version String {} is of invalid format. Must be of format"
               " MajorNum.MinorNum.").format(requested_version)
    resp = render_error_message(
        mv.BAD_REQUEST_STATUS_CODE,
        message,
        mv.BAD_REQUEST_STATUS_NAME,
        max_version=mv.MAX_API_VERSION,
        min_version=mv.MIN_API_VERSION
    )
    flask.abort(resp)


def invalid_param_error(status_code, descr, exc=None):
    LOG.error("Request aborted with status code {code} and "
              "message '{message}'".format(code=status_code, message=descr))

    if exc is not None:
        LOG.error(traceback.format_exc())

    error_code = "INVALID_PARAMS_ON_REQUEST"

    return render_error_message(status_code, descr, error_code)


def internal_error(status_code, descr, exc=None):
    LOG.error("Request aborted with status code {code} and "
              "message '{message}'".format(code=status_code, message=descr))

    if exc is not None:
        LOG.error(traceback.format_exc())

    error_code = "INTERNAL_SERVER_ERROR"
    if status_code == 501:
        error_code = "NOT_IMPLEMENTED_ERROR"

    return render_error_message(status_code, descr, error_code)


def bad_request(error):
    error_code = 400

    LOG.error("Validation Error occurred: "
              "error_code={code}, error_message={message}, "
              "error_name={name}".format(code=error_code,
                                         message=error.message,
                                         name=error.code))

    return render_error_message(error_code, error.message, error.code)


def access_denied(error):
    error_code = 403

    LOG.error("Access Denied: error_code={code}, error_message={message}, "
              "error_name={name}".format(code=error_code,
                                         message=error.message,
                                         name=error.code))

    return render_error_message(error_code, error.message, error.code)


def not_found(error):
    error_code = 404

    LOG.error("Not Found exception occurred: "
              "error_code={code}, error_message={message}, "
              "error_name={name}".format(code=error_code,
                                         message=error.message,
                                         name=error.code))

    return render_error_message(error_code, error.message, error.code)


def to_wrapped_dict(func, id, *args, **kwargs):
    return render(to_wrapped_dict_no_render(func, id, *args, **kwargs))


def to_wrapped_dict_no_render(func, id, *args, **kwargs):
    obj = func(id, *args, **kwargs)
    if obj is None:
        e = ex.NotFoundException(
            {'id': id}, _('Object with %s not found'))

        return not_found(e)
    return obj.to_wrapped_dict()


def _replace_hadoop_version_plugin_version(obj):
    dict.update(obj, {'plugin_version': obj['hadoop_version']})
    dict.pop(obj, 'hadoop_version')


def _replace_tenant_id_project_id(obj):
    dict.update(obj, {'project_id': obj['tenant_id']})
    dict.pop(obj, 'tenant_id')



# Copyright (c) 2017 Massachusetts Open Cloud
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import testtools
from unittest import mock

from sahara import exceptions as ex
from sahara.service.edp.job_binaries.s3.implementation import S3Type
from sahara.tests.unit import base


class TestS3Type(base.SaharaTestCase):

    def setUp(self):
        super(TestS3Type, self).setUp()
        self.i_s = S3Type()

    @mock.patch('sahara.service.edp.job_binaries.s3.implementation.S3Type.'
                'get_raw_data')
    def test_copy_binary_to_cluster(self, get_raw_data):
        remote = mock.Mock()
        job_binary = mock.Mock()
        job_binary.name = 'test'
        job_binary.url = 's3://somebinary'
        get_raw_data.return_value = 'test'

        res = self.i_s.copy_binary_to_cluster(job_binary,
                                              remote=remote)

        self.assertEqual('/tmp/test', res)
        remote.write_file_to.assert_called_with(
            '/tmp/test',
            'test')

    def test_validate_job_location_format(self):
        self.assertTrue(
            self.i_s.validate_job_location_format("s3://temp/temp"))
        self.assertFalse(
            self.i_s.validate_job_location_format("s4://temp/temp"))
        self.assertFalse(self.i_s.validate_job_location_format("s3:///"))

    def test_validate(self):
        data = {"extra": {}, "url": "s3://temp/temp"}
        with testtools.ExpectedException(ex.InvalidDataException):
            self.i_s.validate(data)
        data["extra"] = {"accesskey": "a",
                         "secretkey": "s",
                         "endpoint": "e"}
        self.i_s.validate(data)
        data["extra"].pop("accesskey")
        with testtools.ExpectedException(ex.InvalidDataException):
            self.i_s.validate(data)

    @mock.patch('sahara.service.edp.s3_common.get_raw_job_binary_data')
    def test_get_raw_data(self, s3_get_raw_jbd):
        self.i_s.get_raw_data('a job binary')
        self.assertEqual(1, s3_get_raw_jbd.call_count)



# Copyright (c) 2015 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import mock

from sahara.tests.unit import base
from sahara.utils.openstack import images as sahara_images


class FakeImage(object):
    def __init__(self, name, tags, username):
        self.name = name
        self.tags = tags
        self.username = username


class TestImages(base.SaharaTestCase):
    def setUp(self):
        super(TestImages, self).setUp()
        self.override_config('auth_url', 'https://127.0.0.1:8080/v3/',
                             'trustee')

    @mock.patch('sahara.utils.openstack.base.url_for', return_value='')
    def test_list_registered_images(self, url_for_mock):
        some_images = [
            FakeImage('foo', ['bar', 'baz'], 'test'),
            FakeImage('baz', [], 'test'),
            FakeImage('spam', [], ""),
            FakeImage('spa', [], 'test')]

        with mock.patch(
                'sahara.utils.openstack.images.SaharaImageManager.list',
                return_value=some_images):
            manager = sahara_images.image_manager()

            images = manager.list_registered()
            self.assertEqual(3, len(images))

            images = manager.list_registered(name='foo')
            self.assertEqual(1, len(images))
            self.assertEqual('foo', images[0].name)
            self.assertEqual('test', images[0].username)

            images = manager.list_registered(name='ba')
            self.assertEqual(1, len(images))
            self.assertEqual('baz', images[0].name)
            self.assertEqual('test', images[0].username)

            images = manager.list_registered(name='a')
            self.assertEqual(2, len(images))
            self.assertEqual('baz', images[0].name)
            self.assertEqual('test', images[0].username)
            self.assertEqual('spa', images[1].name)
            self.assertEqual('test', images[1].username)

            images = manager.list_registered(name='eggs')
            self.assertEqual(0, len(images))

            images = manager.list_registered(tags=['bar'])
            self.assertEqual(1, len(images))
            self.assertEqual('foo', images[0].name)

            images = manager.list_registered(tags=['bar', 'eggs'])
            self.assertEqual(0, len(images))

    @mock.patch('sahara.utils.openstack.images.SaharaImageManager.set_meta')
    def test_set_image_info(self, set_meta):
        with mock.patch('sahara.utils.openstack.base.url_for'):
            manager = sahara_images.image_manager()
            manager.set_image_info('id', 'ubuntu')
            self.assertEqual(
                ('id', {'_sahara_username': 'ubuntu'}), set_meta.call_args[0])

            manager.set_image_info('id', 'ubuntu', 'descr')
            self.assertEqual(
                ('id', {'_sahara_description': 'descr',
                        '_sahara_username': 'ubuntu'}),
                set_meta.call_args[0])

    @mock.patch('sahara.utils.openstack.images.SaharaImageManager.get')
    @mock.patch('sahara.utils.openstack.images.SaharaImageManager.delete_meta')
    def test_unset_image_info(self, delete_meta, get_image):
        manager = sahara_images.image_manager()
        image = mock.MagicMock()
        image.tags = ['fake', 'fake_2.0']
        image.username = 'ubuntu'
        image.description = 'some description'
        get_image.return_value = image
        manager.unset_image_info('id')
        self.assertEqual(
            ('id', ['_sahara_tag_fake', '_sahara_tag_fake_2.0',
                    '_sahara_description', '_sahara_username']),
            delete_meta.call_args[0])



# Copyright (c) 2017 Massachusetts Open Cloud
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import testtools
from unittest import mock

from sahara import exceptions as ex
from sahara.service.edp import s3_common
from sahara.tests.unit import base


class FakeJB(object):
    extra = {"accesskey": "access",
             "secretkey": "my-secret",
             "endpoint": "pointy-end"}
    url = "s3://temp/temp"


class S3CommonTestCase(base.SaharaTestCase):

    @mock.patch("botocore.session.Session.create_client")
    @mock.patch("sahara.service.castellan.utils.get_secret")
    def test__get_s3_client(self, cast, boto):
        cast.return_value = "the-actual-password"
        je = FakeJB().extra
        s3_common._get_s3_client(je)
        args = ('s3', None, False, je['endpoint'], je['accesskey'],
                'the-actual-password')
        boto.called_once_with(*args)

    def test__get_names_from_job_binary_url(self):
        self.assertEqual(
            s3_common._get_names_from_job_binary_url("s3://buck"), ["buck"])
        self.assertEqual(
            s3_common._get_names_from_job_binary_url("s3://buck/obj"),
            ["buck", "obj"])
        self.assertEqual(
            s3_common._get_names_from_job_binary_url("s3://buck/dir/obj"),
            ["buck", "dir/obj"])

    def test__get_raw_job_binary_data(self):
        jb = mock.Mock()
        jb.url = "s3://bucket/object"
        boto_conn = mock.Mock()
        boto_conn.head_object = mock.Mock()
        boto_conn.get_object = mock.Mock()
        self.override_config('job_binary_max_KB', 1)

        boto_conn.head_object.return_value = {"ContentLength": 1025}
        self.assertRaises(ex.DataTooBigException,
                          s3_common._get_raw_job_binary_data,
                          jb, boto_conn)

        reader = mock.Mock()
        reader.read = lambda: "the binary"
        boto_conn.get_object.return_value = {"Body": reader}

        boto_conn.head_object.return_value = {"ContentLength": 1024}
        s3_common._get_raw_job_binary_data(jb, boto_conn)

        self.assertEqual(s3_common._get_raw_job_binary_data(jb, boto_conn),
                         "the binary")

        def _raiser():
            raise ValueError
        reader.read = _raiser
        self.assertRaises(ex.S3ClientException,
                          s3_common._get_raw_job_binary_data,
                          jb, boto_conn)

    def test__validate_job_binary_url(self):
        jb_url = "s3://bucket/object"
        s3_common._validate_job_binary_url(jb_url)
        jb_url = "s4://bucket/object"
        with testtools.ExpectedException(ex.BadJobBinaryException):
            s3_common._validate_job_binary_url(jb_url)
        jb_url = "s3://bucket"
        with testtools.ExpectedException(ex.BadJobBinaryException):
            s3_common._validate_job_binary_url(jb_url)

    @mock.patch("sahara.service.edp.s3_common._get_raw_job_binary_data")
    @mock.patch("sahara.service.edp.s3_common._get_s3_client")
    @mock.patch("sahara.service.edp.s3_common._validate_job_binary_url")
    def test_get_raw_job_binary_data(self, validate_jbu, get_s3cl, get_rjbd):
        get_s3cl.return_value = "this would have been boto"
        jb = FakeJB()
        s3_common.get_raw_job_binary_data(jb)
        validate_jbu.assert_called_once_with(jb.url)
        get_s3cl.assert_called_once_with(jb.extra)
        get_rjbd.assert_called_once_with(jb, "this would have been boto")



# Copyright 2015 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import tempfile
from unittest import mock

import jsonschema
from oslo_serialization import jsonutils as json
from oslo_utils import uuidutils

from sahara import context
from sahara.db.templates import api as template_api
from sahara.db.templates import utils as u
from sahara.tests.unit.conductor import base
from sahara.tests.unit.db.templates import common as c

cluster_json = {
    "plugin_name": "vanilla",
    "hadoop_version": "2.7.1",
    "node_groups": [
        {
            "name": "worker",
            "count": 3,
            "node_group_template_id": "{vanilla-260-default-worker}"
        },
        {
            "name": "master",
            "count": 1,
            "node_group_template_id": "{vanilla-260-default-master}"
        }
    ],
    "name": "vanilla-260-default-cluster",
    "neutron_management_network": "{neutron_management_network}",
    "cluster_configs": {}
}

master_json = {
    "plugin_name": "vanilla",
    "hadoop_version": "2.7.1",
    "node_processes": [
        "namenode",
        "resourcemanager",
        "hiveserver"
    ],
    "name": "vanilla-260-default-master",
    "floating_ip_pool": "{floating_ip_pool}",
    "flavor_id": "{flavor_id}",
    "auto_security_group": "{auto_security_group}",
    'security_groups': "{security_groups}"
}

worker_json = {
    "plugin_name": "vanilla",
    "hadoop_version": "2.7.1",
    "node_processes": [
        "nodemanager",
        "datanode"
    ],
    "name": "vanilla-260-default-worker",
    "floating_ip_pool": "{floating_ip_pool}",
    "flavor_id": "{flavor_id}",
    "auto_security_group": "{auto_security_group}",
    'security_groups': "{security_groups}"
}


class Config(c.Config):
    def __init__(self, option_values=None):
        option_values = option_values or {}
        if "name" not in option_values:
            option_values["name"] = "update"
        super(Config, self).__init__(option_values)


class TemplateUpdateTestCase(base.ConductorManagerTestCase):
    def setUp(self):
        super(TemplateUpdateTestCase, self).setUp()
        self.logger = c.Logger()
        template_api.set_logger(self.logger)

    @mock.patch("sahara.utils.api_validator.ApiValidator.validate")
    def test_check_cluster_templates_valid(self, validate):
        self.logger.clear_log()
        ng_templates = [{"template": c.SAMPLE_NGT,
                         "path": "/foo"}]

        # Reference the node group template by name
        clt = copy.copy(c.SAMPLE_CLT)
        clt["node_groups"] = [
            {"name": "test",
             "count": 1,
             "node_group_template_id": "{%s}" % c.SAMPLE_NGT["name"]}
        ]
        cl_templates = [{"template": clt,
                         "path": "/bar"}]

        # Test failed validation
        validate.side_effect = jsonschema.ValidationError("mistake")
        res = template_api.check_cluster_templates_valid(ng_templates,
                                                         cl_templates)
        self.assertTrue(res)
        msg = "Validation for /bar failed, mistake"
        self.assertIn(msg, self.logger.warnings)

        # Validation passes, name replaced
        validate.side_effect = None
        self.logger.clear_log()
        res = template_api.check_cluster_templates_valid(ng_templates,
                                                         cl_templates)
        self.assertFalse(res)
        node_groups = validate.call_args[0][0]["node_groups"]
        self.assertTrue(uuidutils.is_uuid_like(
            node_groups[0]["node_group_template_id"]))

    def test_add_config_section(self):
        # conf here can't be a mock.Mock() because hasattr will
        # return true
        conf = Config()
        conf.register_group = mock.Mock()
        conf.register_opts = mock.Mock()

        template_api.set_conf(conf)
        opts = ["option"]

        # Named config section
        template_api.add_config_section("section", opts)

        self.assertEqual(1, conf.register_group.call_count)
        config_group = conf.register_group.call_args[0][0]
        self.assertEqual("section", config_group.name)
        self.assertEqual([
            mock.call(opts, config_group)], conf.register_opts.call_args_list)

        conf.register_group.reset_mock()
        conf.register_opts.reset_mock()

        # No config section, opts should be registered against
        # the default section
        template_api.add_config_section(None, opts)

        conf.register_group.assert_not_called()
        conf.register_opts.assert_called_with(opts)

    @mock.patch("sahara.db.templates.api.add_config_section")
    def test_add_config_section_for_template(self, add_config_section):

        conf = mock.Mock()
        conf.list_all_sections = mock.Mock()
        template_api.set_conf(conf)

        # No config sections
        conf.list_all_sections.return_value = []
        ngt = c.SAMPLE_NGT
        template_api.add_config_section_for_template(ngt)
        add_config_section.assert_called_with(None,
                                              template_api.all_template_opts)
        add_config_section.reset_mock()

        # Add config section matching plugin
        conf.list_all_sections.return_value += [ngt["plugin_name"]]
        template_api.add_config_section_for_template(ngt)
        add_config_section.assert_called_with(ngt["plugin_name"],
                                              template_api.all_template_opts)
        add_config_section.reset_mock()

        # Add config section matching plugin and version
        section = "{plugin_name}_{hadoop_version}".format(**ngt)
        conf.list_all_sections.return_value += [section]
        template_api.add_config_section_for_template(ngt)
        add_config_section.assert_called_with(section,
                                              template_api.all_template_opts)
        add_config_section.reset_mock()

        # Add config section matching plugin, version and name
        section = "{plugin_name}_{hadoop_version}_{name}".format(**ngt)
        conf.list_all_sections.return_value += [section]
        template_api.add_config_section_for_template(ngt)
        add_config_section.assert_called_with(
            section,
            template_api.node_group_template_opts)
        add_config_section.reset_mock()

        # Add config section matching name
        section = "{name}".format(**ngt)
        conf.list_all_sections.return_value += [section]
        template_api.add_config_section_for_template(ngt)
        add_config_section.assert_called_with(
            section,
            template_api.node_group_template_opts)
        add_config_section.reset_mock()

    def test_substitute_config_values_ngt(self):
        ngt = copy.copy(c.SAMPLE_NGT)
        ngt["flavor_id"] = "{flavor_id}"
        ngt["floating_ip_pool"] = "{floating_ip_pool}"
        configs = {"flavor_id": "2",
                   "floating_ip_pool": None}
        template_api.substitute_config_values(configs, ngt, "/path")
        self.assertEqual("2", ngt["flavor_id"])
        self.assertIsNone(ngt["floating_ip_pool"])

    def test_substitute_config_values_clt(self):
        clt = copy.copy(c.SAMPLE_CLT)
        clt["neutron_management_network"] = "{neutron_management_network}"
        clt["default_image_id"] = "{default_image_id}"

        netid = uuidutils.generate_uuid()
        configs = {"neutron_management_network": netid,
                   "default_image_id": None}
        template_api.substitute_config_values(configs, clt, "/path")
        self.assertEqual(netid, clt["neutron_management_network"])
        self.assertIsNone(clt["default_image_id"])

    def _write_files(self, tempdir, templates):
        files = []
        for template in templates:
            fp = tempfile.NamedTemporaryFile(suffix=".json", mode="w",
                                             dir=tempdir, delete=False)
            json.dump(template, fp)
            files.append(fp.name)
            fp.close()
        return files

    @mock.patch("sahara.db.templates.api.get_configs")
    @mock.patch("sahara.db.templates.api.add_config_section_for_template")
    def test_process_files(self, add_config_section, get_configs):
        self.logger.clear_log()
        tempdir = tempfile.mkdtemp()

        # This should be ignored by process files
        some_other_json = {"name": "fred",
                           "description": "not a template"}

        files = self._write_files(
            tempdir, [cluster_json, master_json, worker_json, some_other_json])

        get_configs.return_value = {"flavor_id": '2', 'security_groups': [],
                                    'auto_security_group': False}
        option_values = {"plugin_name": None,
                         "plugin_version": None}
        template_api.set_conf(Config(option_values))

        # Check that cluster and ng templates are read and returned
        ng_templates, cl_templates = template_api.process_files(tempdir, files)
        cl_temp_names = [f["template"]["name"] for f in cl_templates]
        ng_temp_names = [f["template"]["name"] for f in ng_templates]
        self.assertEqual([cluster_json["name"]], cl_temp_names)
        self.assertEqual([master_json["name"],
                          worker_json["name"]], ng_temp_names)

        # Plugin name/version filtering applied
        option_values = {"plugin_name": "vanilla",
                         "plugin_version": "2.7.1"}
        template_api.set_conf(Config(option_values))
        ng_templates, cl_templates = template_api.process_files(tempdir, files)
        self.assertEqual(1, len(cl_templates))
        self.assertEqual(2, len(ng_templates))

        option_values = {"plugin_name": "hdp",
                         "plugin_version": "2.7.1"}
        template_api.set_conf(Config(option_values))
        ng_templates, cl_templates = template_api.process_files(tempdir, files)
        self.assertEqual(0, len(cl_templates))
        self.assertEqual(0, len(ng_templates))

    @mock.patch("sahara.db.templates.api.get_configs")
    @mock.patch("sahara.db.templates.api.add_config_section_for_template")
    def test_process_files_validation_error(self, add_config_section,
                                            get_configs):
        self.logger.clear_log()
        tempdir = tempfile.mkdtemp()

        files = self._write_files(
            tempdir, [cluster_json, master_json, worker_json])

        get_configs.return_value = {
            "flavor_id": '2',
            'security_groups': [],
            'auto_security_group': False
        }

        option_values = {"plugin_name": None,
                         "plugin_version": None}
        template_api.set_conf(Config(option_values))

        # Bad JSON validation for ng should cause all files to be skipped
        bad_worker = copy.copy(worker_json)
        bad_worker["my_dog"] = ["fido"]
        new_file = self._write_files(tempdir, [bad_worker])[0]
        ng_templates, cl_templates = template_api.process_files(
            tempdir, files + [new_file])
        self.assertEqual(0, len(ng_templates))
        self.assertEqual(0, len(cl_templates))
        msg = ("Validation for {path} failed, "
               "Additional properties are not allowed".format(path=new_file))
        self.assertTrue(self.logger.warnings[0].startswith(msg))

    @mock.patch("sahara.db.templates.api.get_configs")
    @mock.patch("sahara.db.templates.api.add_config_section_for_template")
    def test_process_files_bad_json(self, add_config_section, get_configs):
        self.logger.clear_log()
        tempdir = tempfile.mkdtemp()

        files = self._write_files(
            tempdir, [cluster_json, master_json, worker_json])

        get_configs.return_value = {"flavor_id": '2', 'security_groups': [],
                                    'auto_security_group': False}
        option_values = {"plugin_name": None,
                         "plugin_version": None}
        template_api.set_conf(Config(option_values))

        # Invalid JSON should cause all files to be skipped
        fp = tempfile.NamedTemporaryFile(suffix=".json",
                                         dir=tempdir, delete=False)
        fp.write(b"not json")
        files += [fp.name]
        fp.close()
        ng_templates, cl_templates = template_api.process_files(tempdir, files)
        self.assertEqual(0, len(ng_templates))
        self.assertEqual(0, len(cl_templates))
        msg = ("Error processing {name}".format(name=files[-1]))
        self.assertTrue(self.logger.warnings[0].startswith(msg))
        msg = ("Skipping processing for {dir}, "
               "error processing files".format(dir=tempdir))
        self.assertEqual(msg, self.logger.warnings[1])

    def test_add_node_group_templates(self):
        self.logger.clear_log()
        ctx = context.ctx()

        # Create a record that will be updated in the db
        existing = copy.copy(c.SAMPLE_NGT)
        existing = self.api.node_group_template_create(ctx, existing)

        # Create the update
        update = copy.copy(c.SAMPLE_NGT)
        update["flavor_id"] = "6"

        # Create a record that will be new in the db
        new = copy.copy(c.SAMPLE_NGT)
        new["name"] = "new_name"

        ngts = [{"template": update, "path": "foo"},
                {"template": new, "path": "bar"}]

        ng_info, error = template_api.add_node_group_templates(ctx, ngts)
        self.assertFalse(error)

        new = self.api.node_group_template_get_all(ctx, name=new["name"])[0]
        self.assertIsNotNone(new)

        # ng_info["created"] is a list of templates that were created
        self.assertEqual(1, len(ng_info["created"]))
        self.assertEqual(new["id"], ng_info["created"][0]["id"])

        # ng_info["updated"] is a list of tuples for templates that
        # were updated.  First element in the tuple is the template,
        # second is a dictionary of fields that were updated.
        self.assertEqual(1, len(ng_info["updated"]))
        self.assertEqual(existing["id"], ng_info["updated"][0][0]["id"])
        self.assertEqual({"flavor_id": "42"}, ng_info["updated"][0][1])

        # ng_info["dict"] is a dictionary of name/id pairs
        self.assertEqual({new["name"]: new["id"],
                          existing["name"]: existing["id"]}, ng_info["ids"])

        msg = ("Created node group template {info} from bar".format(
            info=u.name_and_id(new)))
        self.assertIn(msg, self.logger.infos)

        msg = ("Updated node group template {info} from foo".format(
            info=u.name_and_id(existing)))
        self.assertIn(msg, self.logger.infos)

        self.api.node_group_template_destroy(ctx, new["id"],
                                             ignore_prot_on_def=True)
        self.api.node_group_template_destroy(ctx, existing["id"],
                                             ignore_prot_on_def=True)

    @mock.patch("sahara.conductor.API.node_group_template_update")
    @mock.patch("sahara.db.templates.api.reverse_node_group_template_creates")
    @mock.patch("sahara.db.templates.api.reverse_node_group_template_updates")
    def test_add_node_group_templates_update_failed(self,
                                                    reverse_updates,
                                                    reverse_creates,
                                                    ng_update):
        self.logger.clear_log()
        ctx = context.ctx()

        ng_update.side_effect = Exception("mistake")

        # Create a record that will be updated in the db
        existing = copy.copy(c.SAMPLE_NGT)
        existing = self.api.node_group_template_create(ctx, existing)

        # Create the update
        update = copy.copy(c.SAMPLE_NGT)
        update["flavor_id"] = "6"

        # Create a record that will be new in the db
        new = copy.copy(c.SAMPLE_NGT)
        new["name"] = "new_name"

        ngts = [{"template": new, "path": "bar"},
                {"template": update, "path": "foo"}]

        ng_info, error = template_api.add_node_group_templates(ctx, ngts)
        new = self.api.node_group_template_get_all(ctx, name=new["name"])[0]
        self.assertTrue(error)
        self.assertEqual(1, reverse_creates.call_count)

        # call should have been (ctx, [new])
        self.assertEqual(new["id"], reverse_creates.call_args[0][1][0]["id"])

        self.assertEqual(1, reverse_updates.call_count)
        msg = ("Update of node group template {info} failed, mistake".format(
            info=u.name_and_id(existing)))
        self.assertIn(msg, self.logger.warnings)

        self.api.node_group_template_destroy(ctx, new["id"],
                                             ignore_prot_on_def=True)
        self.api.node_group_template_destroy(ctx, existing["id"],
                                             ignore_prot_on_def=True)

    @mock.patch("sahara.conductor.API.node_group_template_create")
    @mock.patch("sahara.db.templates.api.reverse_node_group_template_creates")
    @mock.patch("sahara.db.templates.api.reverse_node_group_template_updates")
    def test_add_node_group_templates_create_failed(self,
                                                    reverse_updates,
                                                    reverse_creates,
                                                    ng_create):
        self.logger.clear_log()
        ctx = context.ctx()

        ng_create.side_effect = Exception("mistake")

        # Create a record that will be updated in the db
        existing = copy.copy(c.SAMPLE_NGT)
        existing = self.api.node_group_template_create(ctx, existing)

        # Create the update
        update = copy.copy(c.SAMPLE_NGT)
        update["flavor_id"] = "6"

        # Create a record that will be new in the db
        new = copy.copy(c.SAMPLE_NGT)
        new["name"] = "new_name"

        ngts = [{"template": update, "path": "foo"},
                {"template": new, "path": "bar"}]

        ng_info, error = template_api.add_node_group_templates(ctx, ngts)
        self.assertTrue(error)
        self.assertEqual(1, reverse_creates.call_count)
        self.assertEqual(1, reverse_updates.call_count)

        # call should have been (ctx, [(existing, updated_fields)])
        self.assertEqual({"flavor_id": existing["flavor_id"]},
                         reverse_updates.call_args[0][1][0][1])

        msg = "Creation of node group template from bar failed, mistake"
        self.assertIn(msg, self.logger.warnings)

        self.api.node_group_template_destroy(ctx, existing["id"],
                                             ignore_prot_on_def=True)

    def test_add_cluster_templates(self):
        self.logger.clear_log()
        ctx = context.ctx()

        # Create a record that will be updated in the db
        existing = copy.copy(c.SAMPLE_CLT)
        existing = self.api.cluster_template_create(ctx, existing)

        # Create the update
        update = copy.copy(c.SAMPLE_CLT)
        update["hadoop_version"] = "1"

        # Create a record that will be new in the db
        new = copy.copy(c.SAMPLE_CLT)
        new["name"] = "new_name"

        clts = [{"template": update, "path": "foo"},
                {"template": new, "path": "bar"}]

        error = template_api.add_cluster_templates(ctx, clts, {})
        self.assertFalse(error)

        new = self.api.cluster_template_get_all(ctx, name=new["name"])[0]
        self.assertIsNotNone(new)

        msg = ("Created cluster template {info} from bar".format(
            info=u.name_and_id(new)))
        self.assertIn(msg, self.logger.infos)

        msg = ("Updated cluster template {info} from foo".format(
            info=u.name_and_id(existing)))
        self.assertIn(msg, self.logger.infos)

        self.api.cluster_template_destroy(ctx, new["id"],
                                          ignore_prot_on_def=True)
        self.api.cluster_template_destroy(ctx, existing["id"],
                                          ignore_prot_on_def=True)

    @mock.patch("sahara.conductor.API.cluster_template_update")
    @mock.patch("sahara.db.templates.api.reverse_cluster_template_creates")
    @mock.patch("sahara.db.templates.api.reverse_cluster_template_updates")
    def test_add_cluster_templates_update_failed(self,
                                                 reverse_updates,
                                                 reverse_creates,
                                                 cl_update):
        self.logger.clear_log()
        ctx = context.ctx()

        cl_update.side_effect = Exception("mistake")

        # Create a record that will be updated in the db
        existing = copy.copy(c.SAMPLE_CLT)
        existing = self.api.cluster_template_create(ctx, existing)

        # Create the update
        update = copy.copy(c.SAMPLE_CLT)
        update["hadoop_version"] = "1"

        # Create a record that will be new in the db
        new = copy.copy(c.SAMPLE_CLT)
        new["name"] = "new_name"

        clts = [{"template": new, "path": "bar"},
                {"template": update, "path": "foo"}]

        error = template_api.add_cluster_templates(ctx, clts, {})
        new = self.api.cluster_template_get_all(ctx, name=new["name"])[0]
        self.assertTrue(error)
        self.assertEqual(1, reverse_creates.call_count)

        # call should have been (ctx, [new])
        self.assertEqual(new["id"], reverse_creates.call_args[0][1][0]["id"])

        self.assertEqual(1, reverse_updates.call_count)
        msg = ("Update of cluster template {info} failed, mistake".format(
            info=u.name_and_id(existing)))
        self.assertIn(msg, self.logger.warnings)

        self.api.cluster_template_destroy(ctx, new["id"],
                                          ignore_prot_on_def=True)
        self.api.cluster_template_destroy(ctx, existing["id"],
                                          ignore_prot_on_def=True)

    @mock.patch("sahara.conductor.API.cluster_template_create")
    @mock.patch("sahara.db.templates.api.reverse_cluster_template_creates")
    @mock.patch("sahara.db.templates.api.reverse_cluster_template_updates")
    def test_add_cluster_templates_create_failed(self,
                                                 reverse_updates,
                                                 reverse_creates,
                                                 cl_create):
        self.logger.clear_log()
        ctx = context.ctx()

        cl_create.side_effect = Exception("mistake")

        # Create a record that will be updated in the db
        existing = copy.copy(c.SAMPLE_CLT)
        existing = self.api.cluster_template_create(ctx, existing)

        # Create the update
        update = copy.copy(c.SAMPLE_CLT)
        update["hadoop_version"] = "1"

        # Create a record that will be new in the db
        new = copy.copy(c.SAMPLE_CLT)
        new["name"] = "new_name"

        clts = [{"template": update, "path": "foo"},
                {"template": new, "path": "bar"}]

        error = template_api.add_cluster_templates(ctx, clts, {})
        self.assertTrue(error)
        self.assertEqual(1, reverse_creates.call_count)
        self.assertEqual(1, reverse_updates.call_count)

        # call should have been (ctx, [(existing, updated_fields)])
        # updated fields will contain hadoop_version and node_groups,
        # since node_groups is modified by the conductor
        updated_fields = reverse_updates.call_args[0][1][0][1]
        self.assertEqual(updated_fields["hadoop_version"],
                         existing["hadoop_version"])
        self.assertIn("node_groups", updated_fields)

        msg = "Creation of cluster template from bar failed, mistake"
        self.assertIn(msg, self.logger.warnings)

        self.api.cluster_template_destroy(ctx, existing["id"],
                                          ignore_prot_on_def=True)

    @mock.patch("sahara.db.templates.api.get_configs")
    @mock.patch("sahara.db.templates.api.add_config_section_for_template")
    def test_do_update_trash(self, add_config, get_configs):
        self.logger.clear_log()
        ctx = context.ctx()

        tempdir = tempfile.mkdtemp()

        self._write_files(tempdir, [cluster_json, master_json, worker_json])

        get_configs.return_value = {
            "flavor_id": '2',
            "neutron_management_network": uuidutils.generate_uuid(),
            'auto_security_group': True,
            'security_groups': [],
        }

        option_values = {"tenant_id": ctx.tenant_id,
                         "directory": tempdir,
                         "norecurse": None,
                         "plugin_name": None,
                         "plugin_version": None}
        template_api.set_conf(Config(option_values))
        template_api.do_update()

        ngs = self.api.node_group_template_get_all(ctx)
        ng_names = sorted([ng["name"] for ng in ngs])
        self.assertEqual(sorted([master_json["name"], worker_json["name"]]),
                         ng_names)

        clts = self.api.cluster_template_get_all(ctx)
        clt_names = sorted([clt["name"] for clt in clts])
        clts = self.api.cluster_template_get_all(ctx)
        self.assertEqual([cluster_json["name"]], clt_names)

    @mock.patch("sahara.db.templates.api.check_cluster_templates_valid")
    @mock.patch("sahara.db.templates.api.get_configs")
    @mock.patch("sahara.db.templates.api.add_config_section_for_template")
    def test_do_update_cluster_invalid(self, add_config,
                                       get_configs, clt_valid):
        self.logger.clear_log()
        ctx = context.ctx()

        tempdir = tempfile.mkdtemp()

        self._write_files(tempdir, [cluster_json, master_json, worker_json])

        get_configs.return_value = {
            "flavor_id": '2',
            "neutron_management_network": uuidutils.generate_uuid()
        }

        option_values = {"tenant_id": ctx.tenant_id,
                         "directory": tempdir,
                         "norecurse": None,
                         "plugin_name": None,
                         "plugin_version": None}
        template_api.set_conf(Config(option_values))

        clt_valid.return_value = True

        template_api.do_update()

        ngs = self.api.node_group_template_get_all(ctx)
        self.assertEqual([], ngs)

        clts = self.api.cluster_template_get_all(ctx)
        self.assertEqual([], clts)

        msg = ("Skipping processing for {dir}, "
               "error processing cluster templates".format(dir=tempdir))
        self.assertIn(msg, self.logger.warnings)

    @mock.patch("sahara.db.templates.api.check_usage_of_existing")
    @mock.patch("sahara.db.templates.api.get_configs")
    @mock.patch("sahara.db.templates.api.add_config_section_for_template")
    def test_do_update_existing_fails(self, add_config,
                                      get_configs, check_existing):
        self.logger.clear_log()
        ctx = context.ctx()

        tempdir = tempfile.mkdtemp()

        self._write_files(tempdir, [cluster_json, master_json, worker_json])

        get_configs.return_value = {
            "flavor_id": '2',
            "neutron_management_network": uuidutils.generate_uuid()
        }
        option_values = {"tenant_id": ctx.tenant_id,
                         "directory": tempdir,
                         "norecurse": None,
                         "plugin_name": None,
                         "plugin_version": None}
        template_api.set_conf(Config(option_values))

        check_existing.return_value = True

        template_api.do_update()

        ngs = self.api.node_group_template_get_all(ctx)
        self.assertEqual([], ngs)

        clts = self.api.cluster_template_get_all(ctx)
        self.assertEqual([], clts)

        msg = ("Skipping processing for {dir}, "
               "templates in use".format(dir=tempdir))
        self.assertIn(msg, self.logger.warnings)

    @mock.patch("sahara.db.templates.api.add_node_group_templates")
    @mock.patch("sahara.db.templates.api.get_configs")
    @mock.patch("sahara.db.templates.api.add_config_section_for_template")
    def test_do_update_add_ngts_fails(self, add_config,
                                      get_configs, add_ngts):
        self.logger.clear_log()
        ctx = context.ctx()

        tempdir = tempfile.mkdtemp()

        self._write_files(tempdir, [cluster_json, master_json, worker_json])

        get_configs.return_value = {
            "flavor_id": '2',
            "neutron_management_network": uuidutils.generate_uuid()
        }

        option_values = {"tenant_id": ctx.tenant_id,
                         "directory": tempdir,
                         "norecurse": None,
                         "plugin_name": None,
                         "plugin_version": None}
        template_api.set_conf(Config(option_values))

        add_ngts.return_value = ({}, True)

        template_api.do_update()

        ngs = self.api.node_group_template_get_all(ctx)
        self.assertEqual([], ngs)

        clts = self.api.cluster_template_get_all(ctx)
        self.assertEqual([], clts)

        msg = ("Skipping processing for {dir}, "
               "error processing node group templates".format(dir=tempdir))
        self.assertIn(msg, self.logger.warnings)

    @mock.patch("sahara.db.templates.api.reverse_node_group_template_creates")
    @mock.patch("sahara.db.templates.api.reverse_node_group_template_updates")
    @mock.patch("sahara.db.templates.api.add_cluster_templates")
    @mock.patch("sahara.db.templates.api.get_configs")
    @mock.patch("sahara.db.templates.api.add_config_section_for_template")
    def test_do_update_add_clts_fails(self,
                                      add_config,
                                      get_configs,
                                      add_clts,
                                      reverse_ng_updates,
                                      reverse_ng_creates):
        self.logger.clear_log()
        ctx = context.ctx()

        tempdir = tempfile.mkdtemp()

        self._write_files(tempdir, [cluster_json, master_json, worker_json])

        get_configs.return_value = {
            "flavor_id": '2',
            "neutron_management_network": uuidutils.generate_uuid()
        }

        option_values = {"tenant_id": ctx.tenant_id,
                         "directory": tempdir,
                         "norecurse": None,
                         "plugin_name": None,
                         "plugin_version": None}
        template_api.set_conf(Config(option_values))

        add_clts.return_value = True

        template_api.do_update()
        self.assertEqual(1, reverse_ng_creates.call_count)
        self.assertEqual(1, reverse_ng_updates.call_count)

        clts = self.api.cluster_template_get_all(ctx)
        self.assertEqual([], clts)

        msg = ("Skipping processing for {dir}, "
               "error processing cluster templates".format(dir=tempdir))
        self.assertIn(msg, self.logger.warnings)



# Copyright (c) 2014 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from sahara.service.edp import base_engine
from sahara.service.validations.edp import job_execution as j
from sahara.utils import edp


class FakeJobEngine(base_engine.JobEngine):
    def cancel_job(self, job_execution):
        pass

    def get_job_status(self, job_execution):
        pass

    def run_job(self, job_execution):
        return 'engine_job_id', edp.JOB_STATUS_SUCCEEDED, None

    def run_scheduled_job(self, job_execution):
        pass

    def validate_job_execution(self, cluster, job, data):
        if job.type == edp.JOB_TYPE_SHELL:
            return
        # All other types except Java require input and output
        # objects and Java require main class
        if job.type in [edp.JOB_TYPE_JAVA, edp.JOB_TYPE_SPARK]:
            j.check_main_class_present(data, job)
        else:
            j.check_data_sources(data, job)

            job_type, subtype = edp.split_job_type(job.type)
            if job_type == edp.JOB_TYPE_MAPREDUCE and (
                    subtype == edp.JOB_SUBTYPE_STREAMING):
                j.check_streaming_present(data, job)

    @staticmethod
    def get_possible_job_config(job_type):
        return None

    @staticmethod
    def get_supported_job_types():
        return edp.JOB_TYPES_ALL



#
# Copyright (c) 2014 Mirantis Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Policy Engine For Sahara"""

import functools

from oslo_config import cfg
from oslo_policy import opts
from oslo_policy import policy

from sahara.common import policies
from sahara import context
from sahara import exceptions

ENFORCER = None

# TODO(gmann): Remove setting the default value of config policy_file
# once oslo_policy change the default value to 'policy.yaml'.
# https://opendev.org/openstack/oslo.policy/src/commit/d8534850d9238e85ae0ea55bf2ac8583681fdb2b/oslo_policy/opts.py#L49
DEFAULT_POLICY_FILE = 'policy.yaml'
opts.set_defaults(cfg.CONF, DEFAULT_POLICY_FILE)


def setup_policy():
    global ENFORCER

    ENFORCER = policy.Enforcer(cfg.CONF)
    ENFORCER.register_defaults(policies.list_rules())


def enforce(rule):
    def decorator(func):
        @functools.wraps(func)
        def handler(*args, **kwargs):
            ctx = context.ctx()
            ENFORCER.authorize(rule, {}, ctx.to_dict(), do_raise=True,
                               exc=exceptions.Forbidden)

            return func(*args, **kwargs)
        return handler

    return decorator



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re

from keystoneauth1.access import service_catalog as keystone_service_catalog
from keystoneauth1 import exceptions as keystone_ex
from oslo_config import cfg
from oslo_log import log as logging
from oslo_serialization import jsonutils as json
from six.moves.urllib import parse as urlparse

from sahara import context
from sahara import exceptions as ex

LOG = logging.getLogger(__name__)

# List of the errors, that can be retried
ERRORS_TO_RETRY = [408, 413, 429, 500, 502, 503, 504]

opts = [
    cfg.IntOpt('retries_number',
               default=5,
               help='Number of times to retry the request to client before '
                    'failing'),
    cfg.IntOpt('retry_after',
               default=10,
               help='Time between the retries to client (in seconds).')
]

retries = cfg.OptGroup(name='retries',
                       title='OpenStack clients calls retries')

CONF = cfg.CONF
CONF.register_group(retries)
CONF.register_opts(opts, group=retries)


def url_for(service_catalog=None, service_type='identity',
            endpoint_type="internalURL"):
    if not service_catalog:
        service_catalog = context.current().service_catalog
    try:
        return keystone_service_catalog.ServiceCatalogV2(
            json.loads(service_catalog)).url_for(
                service_type=service_type, interface=endpoint_type,
                region_name=CONF.os_region_name)
    except keystone_ex.EndpointNotFound:
        return keystone_service_catalog.ServiceCatalogV3(
            json.loads(service_catalog)).url_for(
                service_type=service_type, interface=endpoint_type,
                region_name=CONF.os_region_name)


def prepare_auth_url(auth_url, version):
    info = urlparse.urlparse(auth_url)
    url_path = info.path.rstrip("/")
    # replacing current api version to empty string
    url_path = re.sub('/(v3/auth|v3|v2\.0)', '', url_path)
    url_path = (url_path + "/" + version).lstrip("/")
    return "%s://%s/%s" % (info[:2] + (url_path,))


def retrieve_auth_url(endpoint_type="internalURL", version=None):
    if not version:
        version = 'v3' if CONF.use_identity_api_v3 else 'v2.0'
    ctx = context.current()
    if ctx.service_catalog:
        auth_url = url_for(ctx.service_catalog, 'identity', endpoint_type)
    else:
        auth_url = CONF.trustee.auth_url
    return prepare_auth_url(auth_url, version)


def execute_with_retries(method, *args, **kwargs):
    attempts = CONF.retries.retries_number + 1
    while attempts > 0:
        try:
            return method(*args, **kwargs)
        except Exception as e:
            error_code = getattr(e, 'http_status', None) or getattr(
                e, 'status_code', None) or getattr(e, 'code', None)
            if error_code in ERRORS_TO_RETRY:
                LOG.warning('Occasional error occurred during "{method}" '
                            'execution: {error_msg} ({error_code}). '
                            'Operation will be retried.'.format(
                                method=method.__name__,
                                error_msg=e,
                                error_code=error_code))
                attempts -= 1
                retry_after = getattr(e, 'retry_after', 0)
                context.sleep(max(retry_after, CONF.retries.retry_after))
            else:
                LOG.debug('Permanent error occurred during "{method}" '
                          'execution: {error_msg}.'.format(
                              method=method.__name__, error_msg=e))
                raise e
    else:
        attempts = CONF.retries.retries_number
        raise ex.MaxRetriesExceeded(attempts, method.__name__)



# Copyright 2014 OpenStack Foundation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""add auto_security_groups flag to node group

Revision ID: 010
Revises: 009
Create Date: 2014-07-21 14:31:49.685689

"""

# revision identifiers, used by Alembic.
revision = '010'
down_revision = '009'

from alembic import op
import sqlalchemy as sa

from sahara.db.sqlalchemy import types as st


def upgrade():
    op.add_column('node_group_templates',
                  sa.Column('auto_security_group', sa.Boolean()))
    op.add_column('node_groups',
                  sa.Column('auto_security_group', sa.Boolean()))
    op.add_column('templates_relations',
                  sa.Column('auto_security_group', sa.Boolean()))
    op.add_column('node_groups',
                  sa.Column('open_ports', st.JsonEncoded()))



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
from unittest import mock

from oslo_utils import timeutils

from sahara.conductor import manager
from sahara import context
from sahara.service.castellan import config as castellan
import sahara.service.periodic as p
import sahara.tests.unit.base as base
from sahara.tests.unit.conductor.manager import test_clusters as tc
from sahara.tests.unit.conductor.manager import test_edp as te
from sahara.utils import cluster as c_u


class TestPeriodicBack(base.SaharaWithDbTestCase):

    def setUp(self):
        super(TestPeriodicBack, self).setUp()
        self.api = manager.ConductorManager()
        castellan.validate_config()

    @mock.patch('sahara.service.edp.job_manager.get_job_status')
    def test_job_status_update(self, get_job_status):
        ctx = context.ctx()
        job = self.api.job_create(ctx, te.SAMPLE_JOB)
        ds = self.api.data_source_create(ctx, te.SAMPLE_DATA_SOURCE)
        self._create_job_execution({"end_time": datetime.datetime.now(),
                                    "id": 1},
                                   job, ds, ds)
        self._create_job_execution({"end_time": None,
                                    "id": 2},
                                   job, ds, ds)
        self._create_job_execution({"end_time": None,
                                    "id": 3},
                                   job, ds, ds)
        p._make_periodic_tasks().update_job_statuses(None)
        self.assertEqual(2, get_job_status.call_count)
        get_job_status.assert_has_calls([mock.call('2'),
                                         mock.call('3')])

    @mock.patch('sahara.service.trusts.use_os_admin_auth_token')
    @mock.patch('sahara.service.api.v10.terminate_cluster')
    def test_transient_cluster_terminate(self, terminate_cluster,
                                         use_os_admin_auth_token):

        timeutils.set_time_override(datetime.datetime(2005, 2, 1, 0, 0))

        ctx = context.ctx()
        job = self.api.job_create(ctx, te.SAMPLE_JOB)
        ds = self.api.data_source_create(ctx, te.SAMPLE_DATA_SOURCE)

        self._make_cluster('1')
        self._make_cluster('2')

        self._create_job_execution({"end_time": timeutils.utcnow(),
                                    "id": 1,
                                    "cluster_id": "1"},
                                   job, ds, ds)
        self._create_job_execution({"end_time": None,
                                    "id": 2,
                                    "cluster_id": "2"},
                                   job, ds, ds)
        self._create_job_execution({"end_time": None,
                                    "id": 3,
                                    "cluster_id": "2"},
                                   job, ds, ds)

        timeutils.set_time_override(datetime.datetime(2005, 2, 1, 0, 1))

        p._make_periodic_tasks().terminate_unneeded_transient_clusters(None)
        self.assertEqual(1, terminate_cluster.call_count)
        terminate_cluster.assert_has_calls([mock.call('1')])
        self.assertEqual(1, use_os_admin_auth_token.call_count)

    @mock.patch('sahara.service.api.v10.terminate_cluster')
    def test_not_transient_cluster_does_not_terminate(self, terminate_cluster):

        timeutils.set_time_override(datetime.datetime(2005, 2, 1, 0, 0))
        self._make_cluster('1', is_transient=False)
        timeutils.set_time_override(datetime.datetime(2005, 2, 1, 0, 1))
        p._make_periodic_tasks().terminate_unneeded_transient_clusters(None)

        self.assertEqual(0, terminate_cluster.call_count)

    @mock.patch('sahara.service.api.v10.terminate_cluster')
    def test_transient_cluster_not_killed_too_early(self, terminate_cluster):

        timeutils.set_time_override(datetime.datetime(2005, 2, 1, second=0))

        self._make_cluster('1')

        timeutils.set_time_override(datetime.datetime(2005, 2, 1, second=20))

        p._make_periodic_tasks().terminate_unneeded_transient_clusters(None)
        self.assertEqual(0, terminate_cluster.call_count)

    @mock.patch('sahara.service.trusts.use_os_admin_auth_token')
    @mock.patch('sahara.service.api.v10.terminate_cluster')
    def test_transient_cluster_killed_in_time(self, terminate_cluster,
                                              use_os_admin_auth_token):

        timeutils.set_time_override(datetime.datetime(2005, 2, 1, second=0))

        self._make_cluster('1')

        timeutils.set_time_override(datetime.datetime(2005, 2, 1, second=40))

        p._make_periodic_tasks().terminate_unneeded_transient_clusters(None)
        self.assertEqual(1, terminate_cluster.call_count)
        terminate_cluster.assert_has_calls([mock.call('1')])
        self.assertEqual(1, use_os_admin_auth_token.call_count)

    @mock.patch('sahara.service.api.v10.terminate_cluster')
    def test_incomplete_cluster_not_killed_too_early(self, terminate_cluster):

        self.override_config('cleanup_time_for_incomplete_clusters', 1)

        timeutils.set_time_override(datetime.datetime(2005, 2, 1, second=0))

        self._make_cluster('1', c_u.CLUSTER_STATUS_SPAWNING)

        timeutils.set_time_override(datetime.datetime(
            2005, 2, 1, minute=59, second=50))

        p._make_periodic_tasks().terminate_incomplete_clusters(None)
        self.assertEqual(0, terminate_cluster.call_count)

    @mock.patch('sahara.service.trusts.use_os_admin_auth_token')
    @mock.patch('sahara.service.api.v10.terminate_cluster')
    def test_incomplete_cluster_killed_in_time(self, terminate_cluster,
                                               use_os_admin_auth_token):

        self.override_config('cleanup_time_for_incomplete_clusters', 1)
        timeutils.set_time_override(datetime.datetime(2005, 2, 1, second=0))

        self._make_cluster('1', c_u.CLUSTER_STATUS_SPAWNING)

        timeutils.set_time_override(datetime.datetime(
            2005, 2, 1, hour=1, second=10))

        p._make_periodic_tasks().terminate_incomplete_clusters(None)
        self.assertEqual(1, terminate_cluster.call_count)
        terminate_cluster.assert_has_calls([mock.call('1')])
        self.assertEqual(1, use_os_admin_auth_token.call_count)

    @mock.patch('sahara.service.api.v10.terminate_cluster')
    def test_active_cluster_not_killed_as_inactive(
            self, terminate_cluster):
        self.override_config('cleanup_time_for_incomplete_clusters', 1)

        timeutils.set_time_override(datetime.datetime(2005, 2, 1, second=0))

        self._make_cluster('1')

        timeutils.set_time_override(datetime.datetime(
            2005, 2, 1, hour=1, second=10))

        p._make_periodic_tasks().terminate_incomplete_clusters(None)
        self.assertEqual(0, terminate_cluster.call_count)

    @mock.patch("sahara.utils.proxy.proxy_domain_users_list")
    @mock.patch("sahara.utils.proxy.proxy_user_delete")
    @mock.patch("sahara.service.periodic.conductor.job_execution_get")
    def test_check_for_zombie_proxy_users(self, mock_conductor_je_get,
                                          mock_user_delete,
                                          mock_users_list):
        user_0 = mock.MagicMock()
        user_0.name = "admin"
        user_0.id = 0

        user_1 = mock.MagicMock()
        user_1.name = "job_0"
        user_1.id = 1

        user_2 = mock.MagicMock()
        user_2.name = "job_1"
        user_2.id = 2

        mock_users_list.return_value = [user_0, user_1, user_2]

        je_0 = mock.MagicMock()
        je_0.id = 0
        je_0.info = {"status": "KILLED"}

        je_1 = mock.MagicMock()
        je_1.id = 1
        je_1.info = {"status": "WAITING"}

        mock_conductor_je_get.side_effect = [je_0, je_1]

        p._make_periodic_tasks().check_for_zombie_proxy_users(None)

        mock_user_delete.assert_called_once_with(user_id=1)

    @mock.patch(
        'sahara.service.health.verification_base.validate_verification_start')
    @mock.patch('sahara.service.api.v10.update_cluster')
    def test_run_verifications_executed(self, cluster_update, ver_valid):
        self._make_cluster('1')
        p._make_periodic_tasks().run_verifications(None)
        self.assertEqual(1, ver_valid.call_count)
        cluster_update.assert_called_once_with(
            '1', {'verification': {'status': 'START'}})

    @mock.patch(
        'sahara.service.health.verification_base.validate_verification_start')
    @mock.patch('sahara.service.api.v10.update_cluster')
    def test_run_verifications_not_executed(self, cluster_update, ver_valid):
        self._make_cluster('1', status=c_u.CLUSTER_STATUS_ERROR)
        p._make_periodic_tasks().run_verifications(None)
        ver_valid.assert_not_called()
        cluster_update.assert_not_called()

    @mock.patch("sahara.service.periodic.threadgroup")
    @mock.patch("sahara.service.periodic.CONF")
    def test_setup_enabled(self, mock_conf, mock_thread_group):
        mock_conf.periodic_enable = True
        mock_conf.periodic_fuzzy_delay = 20
        mock_conf.periodic_interval_max = 30
        mock_conf.periodic_workers_number = 1
        mock_conf.periodic_coordinator_backend_url = ''

        add_timer = mock_thread_group.ThreadGroup().add_dynamic_timer

        p.setup()

        self.assertTrue(add_timer._mock_called)

    @mock.patch("sahara.service.periodic.threadgroup")
    @mock.patch("sahara.service.periodic.CONF")
    def test_setup_disabled(self, mock_conf, mock_thread_group):
        mock_conf.periodic_enable = False

        add_timer = mock_thread_group.ThreadGroup().add_dynamic_timer

        p.setup()

        self.assertFalse(add_timer._mock_called)

    def _make_cluster(self, id_name, status=c_u.CLUSTER_STATUS_ACTIVE,

                      is_transient=True):
        ctx = context.ctx()

        c = tc.SAMPLE_CLUSTER.copy()
        c["is_transient"] = is_transient
        c["status"] = status
        c["id"] = id_name
        c["name"] = id_name
        c['updated_at'] = timeutils.utcnow()
        c['trust_id'] = 'DEADBEEF-DEAD-BEEF-DEAD-BEEFDEADBEEF'
        self.api.cluster_create(ctx, c)

    def _create_job_execution(self, values, job, input, output):
        values.update({"job_id": job['id'],
                       "input_id": input['id'],
                       "output_id": output['id']})
        self.api.job_execution_create(context.ctx(), values)
